from gtfspy.routing.label import LabelTimeSimple


class NodeProfileSimple:
    """
    In the connection scan algorithm, each stop has a profile entry
    that stores information on the Pareto-Optimal (departure_time_this_node, arrival_time_target_node) tuples.
    """

    def __init__(self, walk_to_target_duration=float('inf'), label_class=LabelTimeSimple):
        self._labels = []  # list[LabelTimeSimple] # always ordered by decreasing departure_time
        self._walk_to_target_duration = walk_to_target_duration
        self._label_class = label_class

    def get_walk_to_target_duration(self):
        return self._walk_to_target_duration

    def update_pareto_optimal_tuples(self, new_pareto_tuple):
        """
        # this function should be optimized

        Parameters
        ----------
        new_pareto_tuple: LabelTimeSimple

        Returns
        -------
        added: bool
            whether new_pareto_tuple was added to the set of pareto-optimal tuples
        """
        if new_pareto_tuple.duration() > self._walk_to_target_duration:
            direct_walk_label = self._label_class.direct_walk_label(new_pareto_tuple.departure_time,
                                                                    self._walk_to_target_duration)
            if not direct_walk_label.dominates(new_pareto_tuple):
                raise
        direct_walk_label = self._label_class.direct_walk_label(new_pareto_tuple.departure_time, self._walk_to_target_duration)
        if direct_walk_label.dominates(new_pareto_tuple):
            return False

        if self._new_paretotuple_is_dominated_by_old_tuples(new_pareto_tuple):
            return False
        else:
            self._remove_old_tuples_dominated_by_new_and_insert_new_paretotuple(new_pareto_tuple)
            return True

    def _remove_old_tuples_dominated_by_new_and_insert_new_paretotuple(self, new_pareto_tuple):
        indices_to_remove = []
        n = len(self._labels)
        insert_location = 0  # default for the case where len(self._pareto_tuples) == 0
        for i in range(n - 1, -1, -1):
            old_tuple = self._labels[i]
            if old_tuple.departure_time > new_pareto_tuple.departure_time:
                insert_location = i + 1
                break
            else:
                if new_pareto_tuple.dominates(old_tuple):
                    indices_to_remove.append(i)
        for ix in indices_to_remove:
            del self._labels[ix]
        self._labels.insert(insert_location, new_pareto_tuple)

    def _new_paretotuple_is_dominated_by_old_tuples(self, new_pareto_tuple):
        # self._pareto_tuples is guaranteed to be ordered in decreasing dep_time
        # new_pareto_tuple can only be dominated by those elements, which have
        # larger or equal dep_time than new_pareto_tuple.
        for old_tuple in self._labels[::-1]:
            if old_tuple.departure_time >= new_pareto_tuple.arrival_time_target:
                # all following arrival times are necessarily larger
                # than new_pareto_tuple's arrival time
                break
            else:
                if old_tuple.dominates(new_pareto_tuple):
                    return True
        return False

    def evaluate_earliest_arrival_time_at_target(self, dep_time, transfer_margin):
        """
        Get the earliest arrival time at the target, given a departure time.

        Parameters
        ----------
        dep_time : float, int
            time in unix seconds
        transfer_margin: float, int
            transfer margin in seconds

        Returns
        -------
        arrival_time : float
            Arrival time in the given time unit (seconds after unix epoch).
        """
        minimum = dep_time + self._walk_to_target_duration
        dep_time_plus_transfer_margin = dep_time + transfer_margin
        for label in self._labels:
            if label.departure_time >= dep_time_plus_transfer_margin and label.arrival_time_target < minimum:
                minimum = label.arrival_time_target
        return float(minimum)

    def get_final_optimal_labels(self):
        return [label.get_copy() for label in self._labels]




import sqlite3
import pandas as pd


class TravelImpedanceDataStore:

    def __init__(self, db_fname, timeout=100):
        self.db_fname = db_fname
        self.timeout = timeout
        self.conn = sqlite3.connect(self.db_fname, timeout)

    def read_data_as_dataframe(self,
                               travel_impedance_measure,
                               from_stop_I=None,
                               to_stop_I=None,
                               statistic=None):
        """
        Recover pre-computed travel_impedance between od-pairs from the database.

        Returns
        -------
        values: number | Pandas DataFrame
        """
        to_select = []
        where_clauses = []
        to_select.append("from_stop_I")
        to_select.append("to_stop_I")
        if from_stop_I is not None:
            where_clauses.append("from_stop_I=" + str(int(from_stop_I)))
        if to_stop_I is not None:
            where_clauses.append("to_stop_I=" + str(int(to_stop_I)))
        where_clause = ""
        if len(where_clauses) > 0:
            where_clause = " WHERE " + " AND ".join(where_clauses)
        if not statistic:
            to_select.extend(["min", "mean", "median", "max"])
        else:
            to_select.append(statistic)
        to_select_clause = ",".join(to_select)
        if not to_select_clause:
            to_select_clause = "*"
        sql = "SELECT " + to_select_clause + " FROM " + travel_impedance_measure + where_clause + ";"
        df = pd.read_sql(sql, self.conn)
        return df

    def create_table(self, travel_impedance_measure, ensure_uniqueness=True):
        print("Creating table: ", travel_impedance_measure)
        sql = "CREATE TABLE IF NOT EXISTS " + travel_impedance_measure + " (from_stop_I INT, " \
            "to_stop_I INT, " \
            "min REAL, " \
            "max REAL, " \
            "median REAL, " \
            "mean REAL" 
        if ensure_uniqueness:
            sql = sql + ", UNIQUE (from_stop_I, to_stop_I) )"
        else:
            sql = sql + ")"
        self.conn.execute(sql)

    def create_indices_for_all_tables(self, use_memory_as_temp_store=False):
        if use_memory_as_temp_store:
            self.conn.execute("PRAGMA temp_store=2")
        table_names = self.conn.execute("SELECT name FROM sqlite_master WHERE type='table';")
        for table_name in table_names:
            print("Creating indices for table " + str(table_name[0]))
            self.create_indices(table_name[0])
        self.conn.execute("VACUUM")
        self.conn.execute("ANALYZE")

    def create_indices(self, travel_impedance_measure_name):
        table = travel_impedance_measure_name
        sql_from_to = "CREATE UNIQUE INDEX IF NOT EXISTS " + table + "_from_stop_I_to_stop_I ON " + table + " (from_stop_I, to_stop_I)" 
        sql_from = "CREATE INDEX IF NOT EXISTS " + table + "_from_stop_I ON " + table + " (from_stop_I)"
        sql_to = "CREATE INDEX IF NOT EXISTS " + table + "_to_stop_I ON " + table + " (to_stop_I)"
        print("Executing: " + sql_from_to)
        self.conn.execute(sql_from_to)
        print("Executing: " + sql_from)
        self.conn.execute(sql_from)
        print("Executing: " + sql_to)
        self.conn.execute(sql_to)
        self.conn.commit()

    def insert_data(self, travel_impedance_measure_name, data):
        """
        Parameters
        ----------
        travel_impedance_measure_name: str
        data: list[dict]
            Each list element must contain keys:
            "from_stop_I", "to_stop_I", "min", "max", "median" and "mean"
        """
        f = float
        data_tuple = [(int(x["from_stop_I"]), int(x["to_stop_I"]), f(x["min"]), f(x["max"]), f(x["median"]), f(x["mean"])) for
                      x in data]
        insert_stmt = '''INSERT OR REPLACE INTO ''' + travel_impedance_measure_name + ''' (
                              from_stop_I,
                              to_stop_I,
                              min,
                              max,
                              median,
                              mean) VALUES (?, ?, ?, ?, ?, ?) '''
        self.conn.executemany(insert_stmt, data_tuple)
        self.conn.commit()

    def apply_insertion_speedups(self):
        self.conn.execute("PRAGMA SYNCHRONOUS = OFF")






cdef class LabelTimeSimple:
    """
    LabelTime describes entries in a Profile.
    """
    cdef:
        public double departure_time
        public double arrival_time_target

    def __init__(self, double departure_time=-float("inf"), double arrival_time_target=float('inf')):
        self.departure_time = departure_time
        self.arrival_time_target = arrival_time_target

    def __richcmp__(LabelTimeSimple self, LabelTimeSimple other, int op):
        self_tuple = self.departure_time, -self.arrival_time_target
        other_tuple = other.departure_time, -other.arrival_time_target
        if op == 2:  # ==
            return self_tuple == other_tuple
        if op == 3:  # !=
            return self_tuple != other_tuple
        if op == 0:  # less than
            return self_tuple < other_tuple
        elif op == 4:  # greater than
            return self_tuple > other_tuple
        elif op == 1:  # <=
            return self_tuple <= other_tuple
        elif op == 5:  # >=
            return self_tuple >= other_tuple

    # __getstate__ and __setstate__ : for pickling
    def __getstate__(self):
        return self.departure_time, self.arrival_time_target

    def __setstate__(self, state):
        self.departure_time, self.arrival_time_target = state

    cpdef int dominates(self, LabelTimeSimple other) except *:
        return self.departure_time >= other.departure_time and self.arrival_time_target <= other.arrival_time_target

    cpdef int dominates_ignoring_dep_time(self, LabelTimeSimple other):
        return self.arrival_time_target <= other.arrival_time_target

    @staticmethod
    def direct_walk_label(double departure_time, double walk_duration):
        return LabelTimeSimple(departure_time, departure_time + walk_duration)

    cpdef double duration(self):
        return self.arrival_time_target - self.departure_time

    cpdef LabelTimeSimple get_copy(self):
        return LabelTimeSimple(self.departure_time, self.arrival_time_target)

cdef class LabelTime:
    """
    LabelTime describes entries in a Profile.
    """
    cdef:
        public double departure_time
        public double arrival_time_target
        public bint first_leg_is_walk

    def __init__(self, double departure_time=-float("inf"), double arrival_time_target=float('inf'), bint first_leg_is_walk=False,
                 **kwargs):
        self.departure_time = departure_time
        self.arrival_time_target = arrival_time_target
        self.first_leg_is_walk = first_leg_is_walk

    cpdef _tuple_for_ordering(self):
        return self.departure_time, -self.arrival_time_target, not self.first_leg_is_walk

    def __richcmp__(LabelTime self, LabelTime other, int op):
        self_tuple = self._tuple_for_ordering()
        other_tuple = other._tuple_for_ordering()
        if op == 2:  # ==
            return self_tuple == other_tuple
        if op == 3:  # !=
            return self_tuple != other_tuple
        if op == 0:  # less than
            return self_tuple < other_tuple
        elif op == 4:  # greater than
            return self_tuple > other_tuple
        elif op == 1:  # <=
            return self_tuple <= other_tuple
        elif op == 5:  # >=
            return self_tuple >= other_tuple

    # __getstate__ and __setstate__ : for pickling
    def __getstate__(self):
        return self.departure_time, self.arrival_time_target, self.first_leg_is_walk

    def __setstate__(self, state):
        self.departure_time, self.arrival_time_target, self.first_leg_is_walk = state

    cpdef int dominates(LabelTime self, LabelTime other) except *:
        self_tuple = self._tuple_for_ordering()
        other_tuple = other._tuple_for_ordering()
        return all([(s >= o) for s, o in zip(self_tuple, other_tuple)])

    cpdef int dominates_ignoring_dep_time(LabelTime self, LabelTime other):
        return self.arrival_time_target <= other.arrival_time_target and self.first_leg_is_walk <= other.first_leg_is_walk

    cpdef int dominates_ignoring_dep_time_finalization(self, LabelTime other):
        return self.arrival_time_target <= other.arrival_time_target

    cpdef LabelTime get_copy(self):
        return LabelTime(self.departure_time, self.arrival_time_target, self.first_leg_is_walk)

    cpdef LabelTime get_copy_with_specified_departure_time(self, double departure_time):
        return LabelTime(departure_time, self.arrival_time_target, self.first_leg_is_walk)

    @staticmethod
    def direct_walk_label(double departure_time, double walk_duration):
        return LabelTime(departure_time, departure_time + walk_duration, True)

    cpdef double duration(self):
        return self.arrival_time_target - self.departure_time

    cpdef LabelTime get_copy_with_walk_added(self, walk_duration):
        return LabelTime(self.departure_time - walk_duration, self.arrival_time_target, self.first_leg_is_walk)

cdef class LabelTimeWithBoardingsCount:
    cdef:
        public double departure_time
        public double arrival_time_target
        public int n_boardings
        public bint first_leg_is_walk


    def __init__(self, double departure_time, double arrival_time_target,
                 int n_boardings, bint first_leg_is_walk):
        self.departure_time = departure_time
        self.arrival_time_target = arrival_time_target
        self.n_boardings = n_boardings
        self.first_leg_is_walk = first_leg_is_walk

    def __getstate__(self):
        return self.departure_time, self.arrival_time_target, self.n_boardings

    def __setstate__(self, state):
        self.departure_time, self.arrival_time_target, self.n_boardings = state

    def _tuple_for_ordering(self):
        return self.departure_time, -self.arrival_time_target, -self.n_boardings, not self.first_leg_is_walk

    def __richcmp__(LabelTimeWithBoardingsCount self, LabelTimeWithBoardingsCount other, int op):
        self_tuple = self._tuple_for_ordering()
        other_tuple = other._tuple_for_ordering()
        if op == 2:  # ==
            return self_tuple == other_tuple
        if op == 3:  # !=
            return self_tuple != other_tuple
        if op == 0:  # less than
            return self_tuple < other_tuple
        elif op == 4:  # greater than
            return self_tuple > other_tuple
        elif op == 1:  # <=
            return self_tuple <= other_tuple
        elif op == 5:  # >=
            return self_tuple >= other_tuple

    cpdef int dominates(self, LabelTimeWithBoardingsCount other):
        """
        Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles

        Parameters
        ----------
        other: LabelTimeWithBoardingsCount

        Returns
        -------
        dominates: bint
            True if this ParetoTuple dominates the other, otherwise False
        """
        self_tuple = self._tuple_for_ordering()
        other_tuple = other._tuple_for_ordering()
        return not any([(s < o) for s, o in zip(self_tuple, other_tuple)])

    cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeWithBoardingsCount other):
        dominates = (
            self.arrival_time_target <= other.arrival_time_target and
            self.n_boardings <= other.n_boardings
        )
        return dominates

    cpdef int dominates_ignoring_dep_time(self, LabelTimeWithBoardingsCount other):
        cdef:
            int dominates
        dominates = (
            self.arrival_time_target <= other.arrival_time_target and
            self.n_boardings <= other.n_boardings and
            self.first_leg_is_walk <= other.first_leg_is_walk
        )
        return dominates

    cpdef int dominates_ignoring_time(self, LabelTimeWithBoardingsCount other):
        cdef:
            int dominates
        dominates = (
            self.n_boardings <= other.n_boardings and
            self.first_leg_is_walk <= other.first_leg_is_walk
        )
        return dominates

    cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeWithBoardingsCount other):
        cdef:
            int dominates
        dominates = (
            self.arrival_time_target <= other.arrival_time_target and
            self.first_leg_is_walk <= other.first_leg_is_walk
        )
        return dominates

    cpdef get_copy(self):
        return LabelTimeWithBoardingsCount(self.departure_time, self.arrival_time_target,
                                           self.n_boardings, self.first_leg_is_walk)

    cpdef get_copy_with_specified_departure_time(self, departure_time):
        return LabelTimeWithBoardingsCount(departure_time, self.arrival_time_target,
                                           self.n_boardings, self.first_leg_is_walk)

    cpdef double duration(self):
        return self.arrival_time_target - self.departure_time

    @staticmethod
    def direct_walk_label(departure_time, walk_duration):
        return LabelTimeWithBoardingsCount(departure_time, departure_time + walk_duration, 0, True)

    cpdef LabelTimeWithBoardingsCount get_copy_with_walk_added(self, double walk_duration):
        return LabelTimeWithBoardingsCount(self.departure_time - walk_duration,
                                           self.arrival_time_target, self.n_boardings, True)

    def __str__(self):
        return str((self.departure_time, self.arrival_time_target, self.n_boardings, self.first_leg_is_walk))


cdef class LabelVehLegCount:
    cdef:
        public double departure_time
        public int n_boardings
        public bint first_leg_is_walk

    def __init__(self, n_boardings=0, double departure_time=-float('inf'), first_leg_is_walk=False, **kwargs):
        self.n_boardings = n_boardings
        self.departure_time = departure_time
        self.first_leg_is_walk = first_leg_is_walk

    def __getstate__(self):
        return self.departure_time, self.n_boardings, self.first_leg_is_walk

    def __setstate__(self, state):
        self.departure_time, self.n_boardings, self.first_leg_is_walk = state

    def _tuple_for_ordering(self):
        return -self.n_boardings, self.departure_time, not self.first_leg_is_walk

    def __richcmp__(LabelVehLegCount self, LabelVehLegCount other, int op):
        self_tuple = self._tuple_for_ordering()
        other_tuple = other._tuple_for_ordering()
        if op == 2:
            return self_tuple == other_tuple
        if op == 3:
            return self_tuple != other_tuple
        if op == 0:  # less than
            return self_tuple < other_tuple
        elif op == 4:  # greater than
            return self_tuple > other_tuple
        elif op == 1:  # <=
            return self_tuple <= other_tuple
        elif op == 5:  # >=
            return self_tuple >= other_tuple

    cpdef int dominates(self, LabelVehLegCount other) except *:
        """
        Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles

        Parameters
        ----------
        other: LabelTimeWithBoardingsCount

        Returns
        -------
        dominates: bint
            True if this ParetoTuple dominates the other, otherwise False
        """
        return self.n_boardings <= other.n_boardings and self.first_leg_is_walk <= other.first_leg_is_walk

    cpdef int dominates_ignoring_dep_time(self, LabelVehLegCount other):
        return self.dominates(other)

    cpdef int dominates_ignoring_dep_time_finalization(self, LabelVehLegCount other):
        return self.n_boardings <= other.n_boardings

    def get_copy(self):
        return LabelVehLegCount(self.n_boardings, first_leg_is_walk=self.first_leg_is_walk)

    def get_copy_with_specified_departure_time(self, departure_time):
        return LabelVehLegCount(self.n_boardings, departure_time, self.first_leg_is_walk)

    def get_copy_with_walk_added(self, walk_duration):
        return LabelVehLegCount(self.n_boardings, departure_time=self.departure_time - walk_duration,
                                first_leg_is_walk=True)

    @staticmethod
    def direct_walk_label(departure_time, walk_duration):
        return LabelVehLegCount(0, departure_time, True)

# ctypedef fused label:
#     LabelTime
#     LabelTimeWithBoardingsCount
#     LabelVehLegCount
#
# cpdef int dominates(label first, label other):
#     return first.dominates(other)

def compute_pareto_front_smart(list label_list):
    return compute_pareto_front(label_list)

def compute_pareto_front(list label_list, finalization=False, ignore_n_boardings=False):
    pareto_front = []
    if len(label_list) == 0:
        return pareto_front

    assert(not (finalization and ignore_n_boardings))
    # determine function used for domination:
    label = next(iter(label_list))
    if finalization:
        dominates = label.__class__.dominates_ignoring_dep_time_finalization
    elif ignore_n_boardings:
        dominates = label.__class__.dominates_ignoring_dep_time_and_n_boardings
    else:
        dominates = label.__class__.dominates_ignoring_dep_time


    label_list = list(reversed(sorted(label_list)))  # n log(n)
    # assume only that label_list is sorted by departure time (best last)
    current_best_labels_wo_deptime = []
    for new_label in label_list:  # n times
        is_dominated = False
        for best_label in current_best_labels_wo_deptime:
            # the size of current_best_labels_wo_deptime should remain small
            # the new_label can dominate the old ones only partially
            # check if the new one is dominated by the old ones ->
            if dominates(best_label, new_label):
                is_dominated = True
                break
        if is_dominated:
            continue  # do nothing
        else:
            pareto_front.append(new_label)
            new_best = []
            for old_partial_best in current_best_labels_wo_deptime:
                if not dominates(new_label, old_partial_best):
                    new_best.append(old_partial_best)
            new_best.append(new_label)
            current_best_labels_wo_deptime = new_best
    return pareto_front

def compute_pareto_front_naive(list label_list):
    """
    Computes the Pareto frontier of a given label_list

    Parameters
    ----------
    label_list: list[LabelTime]
        (Or any list of objects, for which a function label.dominates(other) is defined.

    Returns
    -------
    pareto_front: list[LabelTime]
        List of labels that belong to the Pareto front.

    Notes
    -----
    Code adapted from:
    http://stackoverflow.com/questions/32791911/fast-calculation-of-pareto-front-in-python
    """
    # cdef:
    #    list dominated
    #    list pareto_front
    #    list remaining
    #    Label other
    #    Label candidate
    dominated = []
    pareto_front = []
    remaining = label_list
    while remaining:  # (is not empty)
        candidate = remaining[0]
        new_remaining = []
        is_dominated = False
        for other_i in range(1, len(remaining)):
            other = remaining[other_i]
            if candidate.dominates(other):
                dominated.append(other)
            else:
                new_remaining.append(other)
                if other.dominates(candidate):
                    is_dominated = True
        if is_dominated:
            dominated.append(candidate)
        else:
            pareto_front.append(candidate)
        remaining = new_remaining
        # after each round:
        #   remaining contains Labels that are not dominated by any in the pareto_front
        #   dominated contains Labels that are dominated by some other LabelTime
    return pareto_front

def merge_pareto_frontiers(labels, labels_other):
    """
    Merge two pareto frontiers by removing dominated entries.

    Parameters
    ----------
    labels: list[LabelTime]
    labels_other: list[LabelTime]

    Returns
    -------
    pareto_front_merged: list[LabelTime]
    """
    def _get_non_dominated_entries(candidates, possible_dominators, survivor_list=None):
        if survivor_list is None:
            survivor_list = list()
        for candidate in candidates:
            candidate_is_dominated = False
            for dominator in possible_dominators:
                if dominator.dominates(candidate):
                    candidate_is_dominated = True
                    break
            if not candidate_is_dominated:
                survivor_list.append(candidate)
        return survivor_list

    survived = []
    survived = _get_non_dominated_entries(labels, labels_other, survivor_list=survived)
    survived = _get_non_dominated_entries(labels_other, survived, survivor_list=survived)
    return survived

def min_arrival_time_target(label_list):
    if len(label_list) > 0:
        return min(label_list, key=lambda label: label.arrival_time_target).arrival_time_target
    else:
        return float('inf')

def min_n_vehicle_trips(label_list):
    if len(label_list) > 0:
        return min(label_list, key=lambda label: label.n_vehicle_trips).n_vehicle_trips
    else:
        return None

cdef class LabelTimeBoardingsAndRoute:
    # Label with implemented added constraint for cases when two labels are tied:
    # The trip with minimal "movement time" should be chosen = maximizing the waiting time for robustness
    cdef:
        public double departure_time
        public double arrival_time_target
        public int n_boardings
        public int movement_duration
        public bint first_leg_is_walk
        public object previous_label
        public object connection


    def __init__(self, double departure_time, double arrival_time_target,
                 int n_boardings, int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):
        self.departure_time = departure_time
        self.arrival_time_target = arrival_time_target
        self.n_boardings = n_boardings
        self.movement_duration = movement_duration
        self.first_leg_is_walk = first_leg_is_walk
        self.previous_label = previous_label
        self.connection = connection

    def __getstate__(self):
        return self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.connection, self.previous_label

    def __setstate__(self, state):
        self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.connection, self.previous_label = state

    def _tuple_for_ordering(self):
        return self.departure_time, -self.arrival_time_target, -self.n_boardings, not self.first_leg_is_walk, -self.movement_duration

    def __richcmp__(LabelTimeBoardingsAndRoute self, LabelTimeBoardingsAndRoute other, int op):
        self_tuple = self._tuple_for_ordering()
        other_tuple = other._tuple_for_ordering()
        if op == 2:  # ==
            return self_tuple == other_tuple
        if op == 3:  # !=
            return self_tuple != other_tuple
        if op == 0:  # less than
            return self_tuple < other_tuple
        elif op == 4:  # greater than
            return self_tuple > other_tuple
        elif op == 1:  # <=
            return self_tuple <= other_tuple
        elif op == 5:  # >=
            return self_tuple >= other_tuple

    cpdef int dominates(self, LabelTimeBoardingsAndRoute other):
        """
        Compute whether this LabelTimeBoardingsAndRoute dominates the other LabelTimeBoardingsAndRoute
        Parameters
        ----------
        other: LabelTimeBoardingsAndRoute
        Returns
        -------
        dominates: bint
            True if this ParetoTuple dominates the other, otherwise False
        """
        self_tuple = self._tuple_for_ordering()
        other_tuple = other._tuple_for_ordering()
        if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
            return False
        elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
            return False
        else:
            return True

    cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeBoardingsAndRoute other):
        if any([self.arrival_time_target > other.arrival_time_target, self.n_boardings > other.n_boardings]):
            return False
        elif all([self.arrival_time_target == other.arrival_time_target, self.n_boardings == other.n_boardings]) \
                and self.movement_duration > other.movement_duration:
            return False
        else:
            return True

    cpdef int dominates_ignoring_dep_time(self, LabelTimeBoardingsAndRoute other):
        if any([self.arrival_time_target > other.arrival_time_target,
                self.n_boardings > other.n_boardings,
                self.first_leg_is_walk > other.first_leg_is_walk]):
            return False
        elif all([self.arrival_time_target == other.arrival_time_target,
                self.n_boardings == other.n_boardings,
                self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
            return False
        else:
            return True

    cpdef int dominates_ignoring_time(self, LabelTimeBoardingsAndRoute other):
        if any([self.n_boardings > other.n_boardings,
                self.first_leg_is_walk > other.first_leg_is_walk]):
            return False
        elif all([self.n_boardings == other.n_boardings,
                self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
            return False
        else:
            return True

    cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeBoardingsAndRoute other):
        if any([self.arrival_time_target > other.arrival_time_target,
                self.first_leg_is_walk > other.first_leg_is_walk]):
            return False
        elif all([self.arrival_time_target == other.arrival_time_target,
                self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
            return False
        else:
            return True

    cpdef get_label_with_connection_added(self, connection):
        return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
                                           self.n_boardings, self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
    cpdef get_copy(self):
        return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
                                           self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)

    cpdef get_copy_with_specified_departure_time(self, departure_time):
        return LabelTimeBoardingsAndRoute(departure_time, self.arrival_time_target,
                                           self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)

    cpdef double duration(self):
        return self.arrival_time_target - self.departure_time

    @staticmethod
    def direct_walk_label(departure_time, walk_duration):
        return LabelTimeBoardingsAndRoute(departure_time, departure_time + walk_duration, 0, True)

    cpdef LabelTimeBoardingsAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
        return LabelTimeBoardingsAndRoute(self.departure_time - walk_duration,
                                           self.arrival_time_target, self.n_boardings, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)

    def __str__(self):
        return str((self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.previous_label, self.connection))

cdef class LabelTimeAndRoute:
    # implement added constraint for cases when two labels are tied:
    # The trip with minimal "movement time" should be chosen = maximizing the waiting time for robustness
    cdef:
        public double departure_time
        public double arrival_time_target
        public int movement_duration
        public bint first_leg_is_walk
        public object previous_label
        public object connection


    def __init__(self, double departure_time, double arrival_time_target,
                 int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):
        self.departure_time = departure_time
        self.arrival_time_target = arrival_time_target
        self.movement_duration = movement_duration
        self.first_leg_is_walk = first_leg_is_walk
        self.previous_label = previous_label
        self.connection = connection

    def __getstate__(self):
        return self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label

    def __setstate__(self, state):
        self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label = state

    def _tuple_for_ordering(self):
        return self.departure_time, -self.arrival_time_target, not self.first_leg_is_walk, -self.movement_duration

    def __richcmp__(LabelTimeAndRoute self, LabelTimeAndRoute other, int op):
        self_tuple = self._tuple_for_ordering()
        other_tuple = other._tuple_for_ordering()
        if op == 2:  # ==
            return self_tuple == other_tuple
        if op == 3:  # !=
            return self_tuple != other_tuple
        if op == 0:  # less than
            return self_tuple < other_tuple
        elif op == 4:  # greater than
            return self_tuple > other_tuple
        elif op == 1:  # <=
            return self_tuple <= other_tuple
        elif op == 5:  # >=
            return self_tuple >= other_tuple

    cpdef int dominates(self, LabelTimeAndRoute other):
        """
        Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles
        Parameters
        ----------
        other: LabelTimeAndRoute
        Returns
        -------
        dominates: bint
            True if this ParetoTuple dominates the other, otherwise False
        """
        self_tuple = self._tuple_for_ordering()
        other_tuple = other._tuple_for_ordering()
        if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
            return False
        elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
            return False
        else:
            return True

    cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeAndRoute other):
        dominates = (
            self.arrival_time_target <= other.arrival_time_target
        )
        return dominates

    cpdef int dominates_ignoring_dep_time(self, LabelTimeAndRoute other):
        cdef:
            int dominates
        dominates = (
            (self.arrival_time_target <= other.arrival_time_target and
            self.first_leg_is_walk <= other.first_leg_is_walk) or
            (self.arrival_time_target == other.arrival_time_target and
            self.first_leg_is_walk == other.first_leg_is_walk and self.movement_duration <= other.movement_duration)

        )
        return dominates

    cpdef int dominates_ignoring_time(self, LabelTimeAndRoute other):
        cdef:
            int dominates
        dominates = (
            self.movement_duration <= other.movement_duration and
            self.first_leg_is_walk <= other.first_leg_is_walk
        )
        return dominates

    cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeAndRoute other):
        cdef:
            int dominates
        dominates = (
            self.arrival_time_target <= other.arrival_time_target and
            self.first_leg_is_walk <= other.first_leg_is_walk
        )
        return dominates

    cpdef get_label_with_connection_added(self, connection):
        return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
                                           self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
    cpdef get_copy(self):
        return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
                                           self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)

    cpdef get_copy_with_specified_departure_time(self, departure_time):
        return LabelTimeAndRoute(departure_time, self.arrival_time_target,
                                           self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)

    cpdef double duration(self):
        return self.arrival_time_target - self.departure_time

    @staticmethod
    def direct_walk_label(departure_time, walk_duration):
        return LabelTimeAndRoute(departure_time, departure_time + walk_duration, 0, True)

    cpdef LabelTimeAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
        return LabelTimeAndRoute(self.departure_time - walk_duration,
                                           self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)

    def __str__(self):
        return str((self.departure_time, self.arrival_time_target, self.movement_duration, self.first_leg_is_walk, self.previous_label, self.connection))


cdef class LabelGeneric:

    # This class is used only for the analysis stage, when the data has been stored in a database
    cdef public int journey_id,  from_stop_I, to_stop_I, n_boardings, movement_duration, journey_duration, \
        in_vehicle_duration, transfer_wait_duration, walking_duration, pre_journey_wait_fp
    cdef public double departure_time, arrival_time_target
    cdef public list attributes

    def __init__(self, journey_dict, **kwargs):
        self.attributes = []
        for key in journey_dict:
            setattr(self, key, journey_dict[key])
            self.attributes.append(key)
        for key in kwargs:
            setattr(self, key, kwargs[key])
            self.attributes.append(key)

        # Assert that key attributes are present
        assert hasattr(self, "journey_id")
        assert hasattr(self, "from_stop_I")
        assert hasattr(self, "to_stop_I")
        assert hasattr(self, "departure_time")
        assert hasattr(self, "arrival_time_target")

    def __getstate__(self):
        return self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label

    def __setstate__(self, state):
        self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label = state

    def _tuple_for_ordering(self):
        return self.departure_time, -self.arrival_time_target, -self.movement_duration

    def __richcmp__(LabelGeneric self, LabelGeneric other, int op):
        self_tuple = self._tuple_for_ordering()
        other_tuple = other._tuple_for_ordering()
        if op == 2:  # ==
            return self_tuple == other_tuple
        if op == 3:  # !=
            return self_tuple != other_tuple
        if op == 0:  # less than
            return self_tuple < other_tuple
        elif op == 4:  # greater than
            return self_tuple > other_tuple
        elif op == 1:  # <=
            return self_tuple <= other_tuple
        elif op == 5:  # >=
            return self_tuple >= other_tuple

    cpdef int dominates(self, LabelGeneric other):
        """
        Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles
        Parameters
        ----------
        other: LabelGeneric
        Returns
        -------
        dominates: bint
            True if this ParetoTuple dominates the other, otherwise False
        """
        self_tuple = self._tuple_for_ordering()
        other_tuple = other._tuple_for_ordering()
        if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
            return False
        elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
            return False
        else:
            return True

    cpdef int dominates_ignoring_dep_time_finalization(self, LabelGeneric other):
        dominates = (
            self.arrival_time_target <= other.arrival_time_target
        )
        return dominates

    cpdef int dominates_ignoring_dep_time(self, LabelGeneric other):
        cdef:
            int dominates
        dominates = (
            (self.arrival_time_target <= other.arrival_time_target) or
            (self.arrival_time_target == other.arrival_time_target and self.movement_duration <= other.movement_duration)

        )
        return dominates

    cpdef int dominates_ignoring_time(self, LabelGeneric other):
        cdef:
            int dominates
        dominates = (
            self.movement_duration <= other.movement_duration
        )
        return dominates

    cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelGeneric other):
        cdef:
            int dominates
        dominates = (
            self.arrival_time_target <= other.arrival_time_target
        )
        return dominates

    cpdef get_label_with_connection_added(self, connection):
        return LabelGeneric(self.departure_time, self.arrival_time_target,
                                           self.movement_duration, connection=connection, previous_label=self)
    cpdef get_copy(self):
        journey_dict = {key: getattr(self, key) for key in self.attributes}
        return LabelGeneric(journey_dict)
        # self.departure_time, self.arrival_time_target, self.movement_duration, connection=self.connection, previous_label=self.previous_label)

    cpdef get_copy_with_specified_departure_time(self, departure_time):
        # journey_dict = {key: getattr(self, key) for key in self.attributes}
        # journey_dict['departure_time'] = departure_time
        return LabelGeneric(departure_time, self.arrival_time_target,  self.movement_duration, connection=self.connection, previous_label=self.previous_label)

    cpdef double duration(self):
        return self.arrival_time_target - self.departure_time

    @staticmethod
    def direct_walk_label(departure_time, walk_duration):
        return LabelGeneric(departure_time, departure_time + walk_duration, 0, True)

    cpdef LabelGeneric get_copy_with_walk_added(self, double walk_duration, object connection):
        return LabelGeneric(self.departure_time - walk_duration,
                                           self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)

    def __str__(self):
        return str((self.departure_time, self.arrival_time_target, self.movement_duration))




# Route types, as specified in the GTFS reference:
# https://developers.google.com/transit/gtfs/reference#routestxt

WALK = -1
TRAM = 0
SUBWAY = 1
RAIL = 2
BUS = 3
FERRY = 4
CABLE_CAR = 5
GONDOLA = 6
FUNICULAR = 7
AIRCRAFT = 99

ALL_ROUTE_TYPES = {WALK, TRAM, SUBWAY, RAIL, BUS, FERRY, CABLE_CAR, GONDOLA, FUNICULAR}
TRANSIT_ROUTE_TYPES = ALL_ROUTE_TYPES.difference({WALK})

ROUTE_TYPE_TO_ZORDER = {
    WALK: 2,
    BUS: 3,
    RAIL: 4,
    SUBWAY: 5,
    TRAM: 6,
    FERRY: 7,
    CABLE_CAR: 8,
    GONDOLA: 9,
    FUNICULAR: 10,
    AIRCRAFT: 11
}

ROUTE_TYPE_TO_DESCRIPTION = {
    WALK: "Walk, pedestrian travel",
    SUBWAY: "Subway, Metro. Any underground rail system within a metropolitan area.",
    RAIL: "Rail. Used for intercity or long - distance travel.",
    BUS: "Bus. Used for short- and long-distance bus routes.",
    FERRY: "Ferry. Used for short- and long-distance boat service.",
    CABLE_CAR: "Cable car. Used for street-level cable cars "
                           "where the cable runs beneath the car.",
    GONDOLA: "Gondola, Suspended cable car. "
                         "Typically used for aerial cable cars where "
                         "the car is suspended from the cable.",
    FUNICULAR: "Funicular. Any rail system designed for steep inclines."
}

ROUTE_TYPE_TO_SHORT_DESCRIPTION = {
    WALK: "Walk",
    TRAM: "Tram",
    SUBWAY: "Subway",
    RAIL: "Rail",
    BUS: "Bus",
    FERRY: "Ferry",
    CABLE_CAR: "Cable car",
    GONDOLA: "Gondola",
    FUNICULAR: "Funicular",
    AIRCRAFT: "Aircraft"
}

ROUTE_TYPE_TO_LOWERCASE_TAG = {
    WALK: "walk",
    TRAM: "tram",
    SUBWAY: "subway",
    RAIL: "rail",
    BUS: "bus",
    FERRY: "ferry",
    CABLE_CAR: "cablecar",
    GONDOLA: "gondola",
    FUNICULAR: "funicular"
}

# Use these on your own risk!
ROUTE_TYPE_TO_APPROXIMATE_CAPACITY = {
    WALK: None,
    TRAM: 200,
    SUBWAY: 600,
    RAIL: 600,
    BUS: 80,
    FERRY: 200,
    CABLE_CAR: 40,
    GONDOLA: 20,
    FUNICULAR: 20
}

ROUTE_TYPE_TO_COLOR = {
    WALK: "black",
    TRAM: '#33a02c',
    SUBWAY: "#ff7f00",
    RAIL: '#e31a1c',
    BUS: '#1f78b4',
    FERRY: "#ffff99",
    CABLE_CAR: "#6a3d9a",
    GONDOLA: "#b15928",
    FUNICULAR: "#fb9a99",
    AIRCRAFT: "#fb9a99"
}


def route_type_to_color_iterable(type):
    return [ROUTE_TYPE_TO_COLOR[x] for x in type]


def route_type_to_zorder(type):
    return [ROUTE_TYPE_TO_ZORDER[x] for x in type]


from __future__ import print_function

import math
import operator

from geoindex import GeoGridIndex, GeoPoint
from geoindex.geo_grid_index import GEO_HASH_GRID_SIZE

from gtfspy.gtfs import GTFS
from gtfspy.util import wgs84_distance, wgs84_height, wgs84_width

create_stmt = ('CREATE TABLE IF NOT EXISTS main.stop_distances '
               '(from_stop_I INT, '
               ' to_stop_I INT, '
               ' d INT, '
               ' d_walk INT, '
               ' min_transfer_time INT, '
               ' timed_transfer INT, '
               'UNIQUE (from_stop_I, to_stop_I)'
               ')'
               )


def bind_functions(conn):
    conn.create_function("find_distance", 4, wgs84_distance)
    conn.create_function("wgs84_height", 1, wgs84_height)
    conn.create_function("wgs84_width", 2, wgs84_width)


def _get_geo_hash_precision(search_radius_in_km):
    # adapted from geoindex.geo_grid_index
    suggested_precision = None
    for precision, max_size in sorted(GEO_HASH_GRID_SIZE.items(), key=operator.itemgetter(1)):
        # ordering is from smallest grid size to largest:
        if search_radius_in_km < max_size / 2:
            suggested_precision = precision
            break
    if suggested_precision is None:
        raise RuntimeError("GeoHash cannot work with this large search radius (km): " + search_radius_in_km)
    return suggested_precision

def calc_transfers(conn, threshold_meters=1000):
    geohash_precision = _get_geo_hash_precision(threshold_meters / 1000.)
    geo_index = GeoGridIndex(precision=geohash_precision)
    g = GTFS(conn)
    stops = g.get_table("stops")
    stop_geopoints = []
    cursor = conn.cursor()

    for stop in stops.itertuples():
        stop_geopoint = GeoPoint(stop.lat, stop.lon, ref=stop.stop_I)
        geo_index.add_point(stop_geopoint)
        stop_geopoints.append(stop_geopoint)
    for stop_geopoint in stop_geopoints:
        nearby_stop_geopoints = geo_index.get_nearest_points_dirty(stop_geopoint, threshold_meters / 1000.0, "km")
        from_stop_I = int(stop_geopoint.ref)
        from_lat = stop_geopoint.latitude
        from_lon = stop_geopoint.longitude

        to_stop_Is = []
        distances = []
        for nearby_stop_geopoint in nearby_stop_geopoints:
            to_stop_I = int(nearby_stop_geopoint.ref)
            if to_stop_I == from_stop_I:
                continue
            to_lat = nearby_stop_geopoint.latitude
            to_lon = nearby_stop_geopoint.longitude
            distance = math.ceil(wgs84_distance(from_lat, from_lon, to_lat, to_lon))
            if distance <= threshold_meters:
                to_stop_Is.append(to_stop_I)
                distances.append(distance)

        n_pairs = len(to_stop_Is)
        from_stop_Is = [from_stop_I]*n_pairs
        cursor.executemany('INSERT OR REPLACE INTO stop_distances VALUES (?, ?, ?, ?, ?, ?);',
                            zip(from_stop_Is, to_stop_Is, distances, [None]*n_pairs, [None]*n_pairs, [None]*n_pairs))
        cursor.execute('CREATE INDEX IF NOT EXISTS idx_sd_fsid ON stop_distances (from_stop_I);')


def _export_transfers(conn, fname):
    conn = GTFS(conn).conn
    cur = conn.cursor()
    cur.execute('SELECT S1.lat, S1.lon, S2.lat, S2.lon, SD.d '
                'FROM stop_distances SD '
                '  LEFT JOIN stops S1 ON (SD.from_stop_I=S1.stop_I) '
                '  LEFT JOIN stops S2 ON (SD.to_stop_I  =S2.stop_I)')
    f = open(fname, 'w')
    for row in cur:
        print(' '.join(str(x) for x in row), file=f)


def main():
    import sys
    cmd = sys.argv[1]
    if cmd == 'calc':
        dbname = sys.argv[2]
        conn = GTFS(dbname).conn
        calc_transfers(conn)
    elif cmd == 'export':
        _export_transfers(sys.argv[2], sys.argv[3])


if __name__ == "__main__":
    main()



import os
import sqlite3
import unittest

from gtfspy.gtfs import GTFS
from gtfspy.import_gtfs import import_gtfs
from gtfspy.mapviz import plot_route_network_from_gtfs


class TestMapviz(unittest.TestCase):

    def setUp(self):
        self.gtfs_source_dir = os.path.join(os.path.dirname(__file__), "test_data/filter_test_feed")
        self.fname = self.gtfs_source_dir + "/test_gtfs.sqlite"
        self._remove_temporary_files()
        conn = sqlite3.connect(self.fname)
        import_gtfs(self.gtfs_source_dir, conn, preserve_connection=True, print_progress=False)
        self.G = GTFS(conn)

    def _remove_temporary_files(self):
        for fn in [self.fname]:
            if os.path.exists(fn) and os.path.isfile(fn):
                os.remove(fn)

    def tearDown(self):
        self._remove_temporary_files()

    def test_plot_trip_counts_per_day(self):
        # simple "it compiles" tests:
        ax = plot_route_network_from_gtfs(self.G)
        ax = plot_route_network_from_gtfs(self.G, map_style="light_all")
        ax = plot_route_network_from_gtfs(self.G, map_style="dark_all")
        ax = plot_route_network_from_gtfs(self.G, map_style="rastertiles/voyager")
        # for interactive testing
        # from matplotlib import pyplot as plt
        # plt.show()



from gtfspy.import_loaders.table_loader import TableLoader


class MetadataLoader(TableLoader):
    """Table to be used for any type of metadata"""
    fname = None
    table = 'metadata'
    tabledef = '(key TEXT UNIQUE NOT NULL, value BLOB, value2 BLOB)'

    @classmethod
    def index(cls, cur):
        cur.execute('CREATE INDEX IF NOT EXISTS idx_metadata_name '
                    'ON metadata (key)')

    @classmethod
    def copy(cls, conn, **where):
        pass


from gtfspy.gtfs import GTFS
from gtfspy.import_loaders.table_loader import TableLoader, decode_six


class FeedInfoLoader(TableLoader):

    """feed_info.txt: various feed metadata"""
    fname = 'feed_info.txt'
    table = 'feed_info'
    tabledef = ('(feed_publisher_name TEXT, '
                'feed_publisher_url TEXT, '
                'feed_lang TEXT, '
                'feed_start_date TEXT, '
                'feed_end_date TEXT, '
                'feed_version TEXT, '
                'feed_id TEXT) ')

    def gen_rows(self, readers, prefixes):
        for reader, prefix in zip(readers, prefixes):
            for row in reader:
                #print row
                start = row['feed_start_date'] if 'feed_start_date' in row else None
                end   = row['feed_end_date']   if 'feed_end_date'   in row else None
                yield dict(
                    feed_publisher_name = decode_six(row['feed_publisher_name']) if 'feed_publisher_name' in row else None,
                    feed_publisher_url  = decode_six(row['feed_publisher_url'])  if 'feed_publisher_url'  in row else None,
                    feed_lang           = decode_six(row['feed_lang'])           if 'feed_lang'           in row else None,
                    feed_start_date     = '%s-%s-%s'%(start[:4], start[4:6], start[6:8])  if start else None,
                    feed_end_date       = '%s-%s-%s'%(end[:4], end[4:6], end[6:8])        if end   else None,
                    feed_version        = decode_six(row['feed_version'])        if 'feed_version'        in row else None,
                    feed_id             = prefix[:-1] if len(prefix) > 0 else prefix
                )

    def post_import2(self, conn):
        # TODO! Something whould be done with this! Multiple feeds are possible, currently only selects one row for all feeds
        G = GTFS(conn)
        for name in ['feed_publisher_name',
                     'feed_publisher_url',
                     'feed_lang',
                     'feed_start_date',
                     'feed_end_date',
                     'feed_version']:
                value = conn.execute('SELECT %s FROM feed_info' % name).fetchone()[0]
                if value:
                    G.meta['feed_info_' + name] = value


from unittest import TestCase

from gtfspy.routing.profile_block_analyzer import ProfileBlockAnalyzer
from gtfspy.routing.profile_block import ProfileBlock

class TestProfileBlockAnalyzer(TestCase):

    def test_interpolate(self):
        blocks = [ProfileBlock(0, 1, 2, 1), ProfileBlock(1, 2, 2, 2)]
        analyzer =  ProfileBlockAnalyzer(blocks, cutoff_distance=3.0)
        self.assertAlmostEqual(analyzer.interpolate(0.2), 1.8)
        self.assertAlmostEqual(analyzer.interpolate(1-10**-9), 1.)
        self.assertAlmostEqual(analyzer.interpolate(1), 1)
        self.assertAlmostEqual(analyzer.interpolate(1.+10**-9), 2)
        self.assertAlmostEqual(analyzer.interpolate(1.23), 2)
        self.assertAlmostEqual(analyzer.interpolate(2), 2)





from gtfspy.import_loaders.table_loader import TableLoader, decode_six


class TransfersLoader(TableLoader):
    """Loader to calculate transfer distances.

    transfer_type, from GTFS spec:
      0/null: recommended transfer point
      1: timed transfer
      2: minimum amount of time
      3: transfers not possible

    """
    # This loader is special.  calc_transfers creates the table there,
    # too.  We put a tabledef here so that copy() will work.
    fname = 'transfers.txt'
    table = 'transfers'
    # TODO: this is copy-pasted from calc_transfers.
    tabledef = ('(from_stop_I INT, '
                'to_stop_I INT, '
                'transfer_type INT, '
                'min_transfer_time INT'
                ')')
    extra_keys = ['from_stop_I',
                  'to_stop_I',
                  ]
    extra_values = ['(SELECT stop_I FROM stops WHERE stop_id=:_from_stop_id)',
                    '(SELECT stop_I FROM stops WHERE stop_id=:_to_stop_id)',
                    ]

    def gen_rows(self, readers, prefixes):
        for reader, prefix in zip(readers, prefixes):
            for row in reader:
                #print row
                yield dict(
                    _from_stop_id     = prefix + decode_six(row['from_stop_id']).strip(),
                    _to_stop_id       = prefix + decode_six(row['to_stop_id']).strip(),
                    transfer_type     = int(row['transfer_type']),
                    min_transfer_time = int(row['min_transfer_time'])
                                        if ('min_transfer_time' in row
                                        and (row.get('min_transfer_time').strip()) )
                                        else None
            )




class ProfileBlock:

    def __init__(self, start_time, end_time, distance_start, distance_end, **extra_properties):
        self.start_time = start_time
        self.end_time = end_time
        assert(self.start_time < self.end_time)
        self.distance_start = distance_start
        self.distance_end = distance_end
        self.extra_properties = extra_properties

    def area(self):
        return self.width() * self.mean()

    def mean(self):
        return 0.5 * (self.distance_start + self.distance_end)

    def width(self):
        return self.end_time - self.start_time

    def max(self):
        return max(self.distance_start, self.distance_end)

    def min(self):
        return min(self.distance_start, self.distance_end)

    def is_flat(self):
        return self.distance_start == self.distance_end

    def interpolate(self, time):
        p = (time - self.start_time) / (self.end_time - self.start_time)
        return (1-p) * self.distance_start + p * self.distance_end

    def __getitem__(self, extra_property_name):
        return self.extra_properties[extra_property_name]

    def __str__(self):
        parts = []
        parts.append(self.start_time)
        parts.append(self.end_time)
        parts.append(self.distance_start)
        parts.append(self.distance_end)
        parts.append(self.extra_properties)
        return str(parts)


import sys
import pandas

# the following is required when using this module as a script
# (i.e. using the if __name__ == "__main__": part at the end of this file)
from gtfspy.warnings_container import WarningsContainer

if __name__ == '__main__' and __package__ is None:
    # import gtfspy
    __package__ = 'gtfspy'


from gtfspy import route_types
from gtfspy.gtfs import GTFS
from gtfspy.util import wgs84_distance


WARNING_5_OR_MORE_CONSECUTIVE_STOPS_WITH_SAME_TIME = "trip--arr_time -combinations with five or more consecutive stops having same stop time"
WARNING_LONG_TRIP_TIME = "Trip time longer than {MAX_TRIP_TIME} seconds"
WARNING_TRIP_UNREALISTIC_AVERAGE_SPEED = "trips whose average speed is unrealistic relative to travel mode"

MAX_ALLOWED_DISTANCE_BETWEEN_CONSECUTIVE_STOPS = 20000  # meters
WARNING_LONG_STOP_SPACING = "distance between consecutive stops longer than " + str(MAX_ALLOWED_DISTANCE_BETWEEN_CONSECUTIVE_STOPS) + " meters"
MAX_TIME_BETWEEN_STOPS = 1800  # seconds
WARNING_LONG_TRAVEL_TIME_BETWEEN_STOPS = "trip--stop_times-combinations with travel time between consecutive stops longer than " + str(MAX_TIME_BETWEEN_STOPS / 60) + " minutes"

WARNING_STOP_SEQUENCE_ORDER_ERROR = "stop sequence is not in right order"
WARNING_STOP_SEQUENCE_NOT_INCREMENTAL = "stop sequences are not increasing always by one in stop_times"
WARNING_STOP_FAR_AWAY_FROM_FILTER_BOUNDARY = "stop far away from spatial filter boundary"

ALL_WARNINGS = {
    WARNING_LONG_STOP_SPACING,
    WARNING_5_OR_MORE_CONSECUTIVE_STOPS_WITH_SAME_TIME,
    WARNING_LONG_TRIP_TIME,
    WARNING_TRIP_UNREALISTIC_AVERAGE_SPEED,
    WARNING_LONG_TRAVEL_TIME_BETWEEN_STOPS,
    WARNING_STOP_SEQUENCE_ORDER_ERROR,
    WARNING_STOP_SEQUENCE_NOT_INCREMENTAL
}

GTFS_TYPE_TO_MAX_SPEED = {
    route_types.TRAM: 100,
    route_types.SUBWAY: 150,
    route_types.RAIL: 300,
    route_types.BUS: 100,
    route_types.FERRY: 80,
    route_types.CABLE_CAR: 50,
    route_types.GONDOLA: 50,
    route_types.FUNICULAR: 50,
    route_types.AIRCRAFT: 1000
}
MAX_TRIP_TIME = 7200  # seconds

class TimetableValidator(object):

    def __init__(self, gtfs, buffer_params=None):
        """
        Parameters
        ----------
        gtfs: GTFS, or path to a GTFS object
            A GTFS object
        """
        if not isinstance(gtfs, GTFS):
            self.gtfs = GTFS(gtfs)
        else:
            self.gtfs = gtfs
        self.buffer_params = buffer_params
        self.warnings_container = WarningsContainer()

    def validate_and_get_warnings(self):
        """
        Validates/checks a given GTFS feed with respect to a number of different issues.

        The set of warnings that are checked for, can be found in the gtfs_validator.ALL_WARNINGS

        Returns
        -------
        warnings: WarningsContainer
        """
        self.warnings_container.clear()
        self._validate_stops_with_same_stop_time()
        self._validate_speeds_and_trip_times()
        self._validate_stop_spacings()
        self._validate_stop_sequence()
        self._validate_misplaced_stops()
        return self.warnings_container

    def _validate_misplaced_stops(self):
        if self.buffer_params:
            p = self.buffer_params
            center_lat = p['lat']
            center_lon = p['lon']
            buffer_distance = p['buffer_distance'] * 1000 * 1.002 # some error margin for rounding
            for stop_row in self.gtfs.stops().itertuples():
                if buffer_distance < wgs84_distance(center_lat, center_lon, stop_row.lat, stop_row.lon):
                    self.warnings_container.add_warning(WARNING_STOP_FAR_AWAY_FROM_FILTER_BOUNDARY, stop_row)
                    print(WARNING_STOP_FAR_AWAY_FROM_FILTER_BOUNDARY, stop_row)

    def _validate_stops_with_same_stop_time(self):
        n_stops_with_same_time = 5
        # this query returns the trips where there are N or more stops with the same stop time
        rows = self.gtfs.get_cursor().execute(
            'SELECT '
            'trip_I, '
            'arr_time, '
            'N '
            'FROM '
            '(SELECT trip_I, arr_time, count(*) as N FROM stop_times GROUP BY trip_I, arr_time) q1 '
            'WHERE N >= ?', (n_stops_with_same_time,)
        )
        for row in rows:
            self.warnings_container.add_warning(WARNING_5_OR_MORE_CONSECUTIVE_STOPS_WITH_SAME_TIME, row)

    def _validate_stop_spacings(self):
        self.gtfs.conn.create_function("find_distance", 4, wgs84_distance)
        # this query calculates distance and travel time between consecutive stops
        rows = self.gtfs.execute_custom_query(
            'SELECT '
            'q1.trip_I, '
            'type, '
            'q1.stop_I as stop_1, '
            'q2.stop_I as stop_2, '
            'CAST(find_distance(q1.lat, q1.lon, q2.lat, q2.lon) AS INT) as distance, '
            'q2.arr_time_ds - q1.arr_time_ds as traveltime '
            'FROM '
            '(SELECT * FROM stop_times, stops WHERE stop_times.stop_I = stops.stop_I) q1, '
            '(SELECT * FROM stop_times, stops WHERE stop_times.stop_I = stops.stop_I) q2, '
            'trips, '
            'routes '
            'WHERE q1.trip_I = q2.trip_I '
            'AND q1.seq + 1 = q2.seq '
            'AND q1.trip_I = trips.trip_I '
            'AND trips.route_I = routes.route_I ').fetchall()
        for row in rows:
            if row[4] > MAX_ALLOWED_DISTANCE_BETWEEN_CONSECUTIVE_STOPS:
                self.warnings_container.add_warning(WARNING_LONG_STOP_SPACING, row)
            if row[5] > MAX_TIME_BETWEEN_STOPS:
                self.warnings_container.add_warning(WARNING_LONG_TRAVEL_TIME_BETWEEN_STOPS, row)

    def _validate_speeds_and_trip_times(self):
        # These are the mode - feasible speed combinations used here:
        # https://support.google.com/transitpartners/answer/1095482?hl=en
        self.gtfs.conn.create_function("find_distance", 4, wgs84_distance)

        # this query returns the total distance and travel time for each trip calculated for each stop spacing separately
        rows = pandas.read_sql(
            'SELECT '
            'q1.trip_I, '
            'type, '
            'sum(CAST(find_distance(q1.lat, q1.lon, q2.lat, q2.lon) AS INT)) AS total_distance, ' # sum used for getting total
            'sum(q2.arr_time_ds - q1.arr_time_ds) AS total_traveltime, ' # sum used for getting total
            'count(*)' # for getting the total number of stops
            'FROM '
            '   (SELECT * FROM stop_times, stops WHERE stop_times.stop_I = stops.stop_I) q1, '
            '   (SELECT * FROM stop_times, stops WHERE stop_times.stop_I = stops.stop_I) q2, '
            '    trips, '
            '    routes '
                'WHERE q1.trip_I = q2.trip_I AND q1.seq + 1 = q2.seq AND q1.trip_I = trips.trip_I '
                     'AND trips.route_I = routes.route_I GROUP BY q1.trip_I', self.gtfs.conn)

        for row in rows.itertuples():
            avg_velocity_km_per_h = row.total_distance / max(row.total_traveltime, 1) * 3.6
            if avg_velocity_km_per_h > GTFS_TYPE_TO_MAX_SPEED[row.type]:
                self.warnings_container.add_warning(WARNING_TRIP_UNREALISTIC_AVERAGE_SPEED + " (route_type=" + str(row.type) + ")",
                                                    row
                )
            if row.total_traveltime > MAX_TRIP_TIME:
                self.warnings_container.add_warning(WARNING_LONG_TRIP_TIME.format(MAX_TRIP_TIME=MAX_TRIP_TIME), row, 1)

    def _validate_stop_sequence(self):
        # This function checks if the seq values in stop_times are increasing with departure_time,
        # and that seq always increases by one.
        rows = self.gtfs.execute_custom_query('SELECT trip_I, dep_time_ds, seq '
                                              'FROM stop_times '
                                              'ORDER BY trip_I, dep_time_ds, seq').fetchall()
        old_trip_id = None
        old_seq = None
        for row in rows:
            new_trip_id = int(row[0])
            new_seq = int(row[2])
            if old_trip_id == new_trip_id:
                if old_seq + 1 != new_seq:
                    self.warnings_container.add_warning(WARNING_STOP_SEQUENCE_NOT_INCREMENTAL, row)
                if old_seq >= new_seq:
                    self.warnings_container.add_warning(WARNING_STOP_SEQUENCE_ORDER_ERROR, row)
            old_trip_id = row[0]
            old_seq = row[2]


def main():
    cmd = sys.argv[1]
    args = sys.argv[2:]
    if cmd == "validate":
        validator = TimetableValidator(args[0])
        warningscontainer = validator.validate_and_get_warnings()
        warningscontainer.write_summary()

if __name__ == "__main__":
    main()




from gtfspy.routing.connection import Connection


class ForwardJourney:
    """
    A class for handling journeys generated by routing algorithms
    Current assumptions:
    - legs are added in order of travel
    - walking trips do not have a trip_id
    """

    def __init__(self, legs=None):
        """
        Parameters
        ----------
        legs: list[Connection]
        """
        self.legs = []
        self.departure_time = None
        self.arrival_time = None
        self.trip_ids = set()
        self.n_boardings = 0
        if legs is not None:
            for leg in legs:
                self.add_leg(leg)

    def add_leg(self, leg):
        """
        Parameters
        ----------
        leg: Connection
        """
        assert(isinstance(leg, Connection))
        if not self.legs:
            self.departure_time = leg.departure_time
        self.arrival_time = leg.arrival_time
        if leg.trip_id and (not self.legs or (leg.trip_id != self.legs[-1].trip_id)):
            self.n_boardings += 1
        self.arrival_time = leg.arrival_time
        self.legs.append(leg)

    def get_legs(self):
        return self.legs

    def get_travel_time(self):
        travel_time = self.arrival_time - self.departure_time
        return travel_time

    def get_transfers(self):
        return max(self.n_boardings - 1, 0)

    def get_all_stops(self):
        all_stops = []
        for leg in self.legs:
            all_stops.append(leg.departure_stop)
        all_stops.append(self.legs[-1].arrival_stop)
        return all_stops

    def get_transfer_stop_pairs(self):
        """
        Get stop pairs through which transfers take place

        Returns
        -------
        transfer_stop_pairs: list
        """
        transfer_stop_pairs = []
        previous_arrival_stop = None
        current_trip_id = None
        for leg in self.legs:
            if leg.trip_id is not None and leg.trip_id != current_trip_id and previous_arrival_stop is not None:
                transfer_stop_pair = (previous_arrival_stop, leg.departure_stop)
                transfer_stop_pairs.append(transfer_stop_pair)
            previous_arrival_stop = leg.arrival_stop
            current_trip_id = leg.trip_id
        return transfer_stop_pairs

    def get_transfer_trip_pairs(self):
        pass

    def get_waiting_times(self):
        waiting_times = []
        previous_arrival = None
        for leg in self.legs:
            current_departure = leg.departure_time
            if previous_arrival:
                waiting_times.append(current_departure - previous_arrival)
            previous_arrival = leg.arrival_time
        return waiting_times

    def get_total_waiting_time(self):
        waiting_times = self.get_waiting_times()
        waiting_total = sum(waiting_times)
        return waiting_total

    def get_invehicle_times(self):
        invehicle_times = []
        for leg in self.legs:
            assert(isinstance(leg, Connection))
            if leg.trip_id is not None:
                invehicle_times.append(leg.duration())
        return invehicle_times

    def get_total_invehicle_time(self):
        return sum(self.get_invehicle_times())

    def get_walking_times(self):
        walking_times = []
        for leg in self.legs:
            if leg.is_walk:
                walking_times.append(leg.duration() - leg.waiting_time)
        return walking_times

    def get_total_walking_time(self):
        return sum(self.get_walking_times())

    def dominates(self, other, consider_time=True, consider_boardings=True):
        if consider_time:
            dominates_time = (self.departure_time >= other.departure_time and
                              self.arrival_time <= other.arrival_time)
            if not dominates_time:
                return False
        if consider_boardings:
            dominates_boardings = (self.n_boardings <= other.n_boardings)
            if not dominates_boardings:
                return False
        # dominates w.r.t all aspects:
        return True





"""Shape tools

This module contains tools for dealing with shapes.

The core data structure for representing shapes is
[ {'seq':N, 'lat':F, 'lon':F}, ... ].
`seq` is the integer index of sequence points (note that GTFS doesn't
require integers), and lat/lon are latitude/longitude.  Note that
there is no stop_id or other data stored, though any other data can be
put in the dictionary and it will be preserved.  The 'd' key, if
present, represents cumulative distance (m) along the points.

This same structure is used for representing sequences of stops.  The
function `find_segments` take a shape-sequence and stop-sequence, and
returns a list of `break_points`, which are the indexes within
shape-sequence which correspond to each stop within stop-sequence.

You can then pass `break_points` and the shape-sequence to the
function `return_segments`, which returns a list
[ [ shape-sequence ],
  [ shape-sequence ],
  ...,
  [ ]
]
This is the shape, broken down into individual segments, one
corresponding to each stop.  The last one is empty.


"""
from __future__ import print_function
from __future__ import absolute_import

import numpy as np
from .util import wgs84_distance


def print_coords(rows, prefix=''):
    """Print coordinates within a sequence.

    This is only used for debugging.  Printed in a form that can be
    pasted into Python for visualization."""
    lat = [row['lat'] for row in rows]
    lon = [row['lon'] for row in rows]
    print('COORDS'+'-' * 5)
    print("%slat, %slon = %r, %r" % (prefix, prefix, lat, lon))
    print('-'*5)


def find_segments(stops, shape):
    """Find corresponding shape points for a list of stops and create shape break points.

    Parameters
    ----------
    stops: stop-sequence (list)
        List of stop points
    shape: list of shape points
        shape-sequence of shape points

    Returns
    -------
    break_points: list[int]
        stops[i] corresponds to shape[break_points[i]].  This list can
        be used to partition the shape points into segments between
        one stop and the next.
    badness: float
        Lower indicates better fit to the shape.  This is the sum of
        distances (in meters) between every each stop and its closest
        shape point.  This is not needed in normal use, but in the
        cases where you must determine the best-fitting shape for a
        stop-sequence, use this.
    """
    if not shape:
        return [], 0
    break_points = []
    last_i = 0
    cumul_d = 0
    badness = 0
    d_last_stop = float('inf')
    lstlat, lstlon = None, None
    break_shape_points = []
    for stop in stops:
        stlat, stlon = stop['lat'], stop['lon']
        best_d = float('inf')
        # print stop
        if badness > 500 and badness > 30 * len(break_points):
            return [], badness
        for i in range(last_i, len(shape)):
            d = wgs84_distance(stlat, stlon, shape[i]['lat'], shape[i]['lon'])
            if lstlat:
                d_last_stop = wgs84_distance(lstlat, lstlon, shape[i]['lat'], shape[i]['lon'])
            # If we are getting closer to next stop, record this as
            # the best stop so far.continue
            if d < best_d:
                best_d = d
                best_i = i
                # print best_d, i, last_i, len(shape)
                cumul_d += d
            # We have to be very careful about our stop condition.
            # This is trial and error, basically.
            if (d_last_stop < d) or (d > 500) or (i < best_i + 100):
                    continue
            # We have decided our best stop, stop looking and continue
            # the outer loop.
            else:
                badness += best_d
                break_points.append(best_i)
                last_i = best_i
                lstlat, lstlon = stlat, stlon
                break_shape_points.append(shape[best_i])
                break
        else:
            # Executed if we did *not* break the inner loop
            badness += best_d
            break_points.append(best_i)
            last_i = best_i
            lstlat, lstlon = stlat, stlon
            break_shape_points.append(shape[best_i])
            pass
    # print "Badness:", badness
    # print_coords(stops, 'stop')
    # print_coords(shape, 'shape')
    # print_coords(break_shape_points, 'break')
    return break_points, badness


def find_best_segments(cur, stops, shape_ids, route_id=None,
                       breakpoints_cache=None):
    """Finds the best shape_id for a stop-sequence.

    This is used in cases like when you have GPS data with a route
    name, but you don't know the route direction.  It tries shapes
    going both directions and returns the shape that best matches.
    Could be used in other cases as well.

    Parameters
    ----------
    cur : sqlite3.Cursor
        database cursor
    stops : list
    shape_ids : list of shape_id:s
    route_id : route_id to search for stops
    breakpoints_cache : dict
        If given, use this to cache results from this function.
    """
    cache_key = None
    if breakpoints_cache is not None:
        # Calculate a cache key for this sequence.  If shape_id and
        # all stop_Is are the same, then we assume that it is the same
        # route and re-use existing breakpoints.
        cache_key = (route_id, tuple(x['stop_I'] for x in stops))
        if cache_key in breakpoints_cache:
            print('found in cache')
            return breakpoints_cache[cache_key]

    if route_id is not None:
        cur.execute('''SELECT DISTINCT shape_id
                        FROM routes
                        LEFT JOIN trips
                        USING (route_I)
                        WHERE route_id=?''',
                    (route_id,))
        data = cur.fetchall()
        # If not data, then route_id didn't match anything, or there
        # were no shapes defined.  We have to exit in this case.
        if not data:
            print("No data for route_id=%s" % route_id)
            return [], None, None, None
        #
        shape_ids = zip(*data)[0]
    # print 'find_best_segments:', shape_ids
    results = []
    for shape_id in shape_ids:
        shape = get_shape_points(cur, shape_id)
        breakpoints, badness = find_segments(stops, shape)
        results.append([badness, breakpoints, shape, shape_id])
        if len(stops) > 5 and badness < 5*(len(stops)):
            break

    best = np.argmin(zip(*results)[0])
    # print 'best', best
    badness = results[best][0]
    breakpoints = results[best][1]
    shape = results[best][2]
    shape_id = results[best][3]
    if breakpoints_cache is not None:
        print("storing in cache", cache_key[0], hash(cache_key[1:]))
        breakpoints_cache[cache_key] = breakpoints, badness, shape, shape_id
    return breakpoints, badness, shape, shape_id


def return_segments(shape, break_points):
    """Break a shape into segments between stops using break_points.

    This function can use the `break_points` outputs from
    `find_segments`, and cuts the shape-sequence into pieces
    corresponding to each stop.
    """
    # print 'xxx'
    # print stops
    # print shape
    # print break_points
    # assert len(stops) == len(break_points)
    segs = []
    bp = 0 # not used
    bp2 = 0
    for i in range(len(break_points)-1):
        bp = break_points[i] if break_points[i] is not None else bp2
        bp2 = break_points[i+1] if break_points[i+1] is not None else bp
        segs.append(shape[bp:bp2+1])
    segs.append([])
    return segs


def gen_cumulative_distances(stops):
    """
    Add a 'd' key for distances to a stop/shape-sequence.

    This takes a shape-sequence or stop-sequence, and adds an extra
    'd' key that is cumulative, geographic distances between each
    point. This uses `wgs84_distance` from the util module.  The
    distances are in meters.  Distances are rounded to the nearest
    integer, because otherwise JSON size increases greatly.

    Parameters
    ----------
    stops: list
        elements are dicts with 'lat' and 'lon' keys
        and the function adds the 'd' key ('d' stands for distance)
        to the dictionaries
    """
    stops[0]['d'] = 0.0
    for i in range(1, len(stops)):
        stops[i]['d'] = stops[i-1]['d'] + wgs84_distance(
            stops[i-1]['lat'], stops[i-1]['lon'],
            stops[i]['lat'], stops[i]['lon'],
            )
    for stop in stops:
        stop['d'] = int(stop['d'])
        # stop['d'] = round(stop['d'], 1)


def get_shape_points(cur, shape_id):
    """
    Given a shape_id, return its shape-sequence.

    Parameters
    ----------
    cur: sqlite3.Cursor
        cursor to a GTFS database
    shape_id: str
        id of the route

    Returns
    -------
    shape_points: list
        elements are dictionaries containing the 'seq', 'lat', and 'lon' of the shape
    """
    cur.execute('''SELECT seq, lat, lon, d FROM shapes where shape_id=?
                    ORDER BY seq''', (shape_id,))
    shape_points = [dict(seq=row[0], lat=row[1], lon=row[2], d=row[3])
                    for row in cur]
    return shape_points


def get_shape_points2(cur, shape_id):
    """
    Given a shape_id, return its shape-sequence (as a dict of lists).
    get_shape_points function returns them as a list of dicts

    Parameters
    ----------
    cur: sqlite3.Cursor
        cursor to a GTFS database
    shape_id: str
        id of the route

    Returns
    -------
    shape_points: dict of lists
        dict contains keys 'seq', 'lat', 'lon', and 'd'(istance) of the shape
    """
    cur.execute('''SELECT seq, lat, lon, d FROM shapes where shape_id=?
                    ORDER BY seq''', (shape_id,))
    shape_points = {'seqs': [], 'lats':  [], 'lons': [], 'd': []}
    for row in cur:
        shape_points['seqs'].append(row[0])
        shape_points['lats'].append(row[1])
        shape_points['lons'].append(row[2])
        shape_points['d'].append(row[3])
    return shape_points


def get_route_shape_segments(cur, route_id):
    """
    Given a route_id, return its stop-sequence.

    Parameters
    ----------
    cur: sqlite3.Cursor
        cursor to a GTFS database
    route_id: str
        id of the route

    Returns
    -------
    shape_points: list
        elements are dictionaries containing the 'seq', 'lat', and 'lon' of the shape
    """
    cur.execute('''SELECT seq, lat, lon
                    FROM (
                        SELECT shape_id
                        FROM route
                        LEFT JOIN trips
                        USING (route_I)
                        WHERE route_id=? limit 1
                        )
                    JOIN shapes
                    USING (shape_id)
                    ORDER BY seq''', (route_id,))
    shape_points = [dict(seq=row[0], lat=row[1], lon=row[2]) for row in cur]
    return shape_points


def get_shape_between_stops(cur, trip_I, seq_stop1=None, seq_stop2=None, shape_breaks=None):
    """
    Given a trip_I (shortened id), return shape points between two stops
    (seq_stop1 and seq_stop2).

    Trip_I is used for matching obtaining the full shape of one trip (route).
    From the resulting shape we then obtain only shape points between
    stop_seq1 and stop_seq2
    trip_I---(trips)--->shape_id
    trip_I, seq_stop1----(stop_times)---> shape_break1
    trip_I, seq_stop2----(stop_times)---> shape_break2
    shapes_id+shape_break1+shape_break2 --(shapes)--> result

    Parameters
    ----------
    cur : sqlite3.Cursor
        cursor to sqlite3 DB containing GTFS
    trip_I : int
        transformed trip_id (i.e. a new column that is created when
        GTFS is imported to a DB)
    seq_stop1: int
        a positive inger describing the index of the point of the shape that
        corresponds to the first stop
    seq_stop2: int
        a positive inger describing the index of the point of the shape that
        corresponds to the second stop
    shape_breaks: ??

    Returns
    -------
    shapedict: dict
        Dictionary containing the latitudes and longitudes:
            lats=shapedict['lat']
            lons=shapedict['lon']
    """

    assert (seq_stop1 and seq_stop2) or shape_breaks
    if not shape_breaks:
        shape_breaks = []
        for seq_stop in [seq_stop1, seq_stop2]:
            query = """SELECT shape_break FROM stop_times
                        WHERE trip_I=%d AND seq=%d
                    """ % (trip_I, seq_stop)
            for row in cur.execute(query):
                shape_breaks.append(row[0])
    assert len(shape_breaks) == 2

    query = """SELECT seq, lat, lon
                FROM (SELECT shape_id FROM trips WHERE trip_I=%d)
                JOIN shapes USING (shape_id)
                WHERE seq>=%d AND seq <= %d;
            """ % (trip_I, shape_breaks[0], shape_breaks[1])
    shapedict = {'lat': [], 'lon': [], 'seq': []}
    for row in cur.execute(query):
        shapedict['seq'].append(row[0])
        shapedict['lat'].append(row[1])
        shapedict['lon'].append(row[2])
    return shapedict


def get_trip_points(cur, route_id, offset=0, tripid_glob=''):
    """Get all scheduled stops on a particular route_id.

    Given a route_id, return the trip-stop-list with
    latitude/longitudes.  This is a bit more tricky than it seems,
    because we have to go from table route->trips->stop_times.  This
    functions finds an arbitrary trip (in trip table) with this route ID
    and, and then returns all stop points for that trip.

    Parameters
    ----------
    cur : sqlite3.Cursor
        cursor to sqlite3 DB containing GTFS
    route_id : string or any
        route_id to get stop points of
    offset : int
        LIMIT offset if you don't want the first trip returned.
    tripid_glob : string
        If given, allows you to limit tripids which can be selected.
        Mainly useful in debugging.

    Returns
    -------
    stop-list
        List of stops in stop-seq format.
    """
    extra_where = ''
    if tripid_glob:
        extra_where = "AND trip_id GLOB '%s'" % tripid_glob
    cur.execute('SELECT seq, lat, lon '
                'FROM (select trip_I from route '
                '      LEFT JOIN trips USING (route_I) '
                '      WHERE route_id=? %s limit 1 offset ? ) '
                'JOIN stop_times USING (trip_I) '
                'LEFT JOIN stop USING (stop_id) '
                'ORDER BY seq' % extra_where, (route_id, offset))
    stop_points = [dict(seq=row[0], lat=row[1], lon=row[2]) for row in cur]
    return stop_points


def interpolate_shape_times(shape_distances, shape_breaks, stop_times):
    """
    Interpolate passage times for shape points.

    Parameters
    ----------
    shape_distances: list
        list of cumulative distances along the shape
    shape_breaks: list
        list of shape_breaks
    stop_times: list
        list of stop_times

    Returns
    -------
    shape_times: list of ints (seconds) / numpy array
        interpolated shape passage times

    The values of stop times before the first shape-break are given the first
    stopping time, and the any shape points after the last break point are
    given the value of the last shape point.
    """
    shape_times = np.zeros(len(shape_distances))
    shape_times[:shape_breaks[0]] = stop_times[0]
    for i in range(len(shape_breaks)-1):
        cur_break = shape_breaks[i]
        cur_time = stop_times[i]
        next_break = shape_breaks[i+1]
        next_time = stop_times[i+1]
        if cur_break == next_break:
            shape_times[cur_break] = stop_times[i]
        else:
            cur_distances = shape_distances[cur_break:next_break+1]
            norm_distances = ((np.array(cur_distances)-float(cur_distances[0])) /
                              float(cur_distances[-1] - cur_distances[0]))
            times = (1.-norm_distances)*cur_time+norm_distances*next_time
            shape_times[cur_break:next_break] = times[:-1]
    # deal final ones separately:
    shape_times[shape_breaks[-1]:] = stop_times[-1]
    return list(shape_times)



from gtfspy import calc_transfers
from gtfspy.import_loaders.table_loader import TableLoader


class StopDistancesLoader(TableLoader):
    """Loader to calculate transfer distances.

    The stop_distances table has the following columns:
    from_stop_I:        reference stop_I
    to_stop_I:          reference stop_I
    d:                  straight-line distance (M)
    d_walk:             distance, routed somehow
    min_transfer_time:  from transfers.txt, in seconds.
    """
    # This loader is special.  calc_transfers creates the table there,
    # too.  We put a tabledef here so that copy() will work.
    fname = None
    table = 'stop_distances'
    tabledef = calc_transfers.create_stmt
    threshold = 1000

    def post_import(self, cur):
        # why is cur not used?
        conn = self._conn
        cur = conn.cursor()
        cur2 = conn.cursor()
        if self.print_progress:
            print("Calculating straight-line transfer distances")
        calc_transfers.calc_transfers(conn, threshold_meters=self.threshold)

        # Copy data from transfers table.  Several steps below.
        if self.print_progress:
            print("Copying information from transfers to stop_distances.")
        calc_transfers.bind_functions(conn)

        # Add min transfer times (transfer_type=2).  This just copies
        # min_transfer_time from `transfers` to `stop_distances`.
        stmt = ('SELECT min_transfer_time, from_stop_I, to_stop_I '
                'FROM transfers '
                'WHERE transfer_type=2 '
                'and from_stop_I!=to_stop_I')
        # First we have to run with INSERT OR IGNORE to add in any
        # rows that are missing.  Unfortunately there is no INSERT OR
        # UPDATE, so we do this in two stages.  First we insert any
        # missing rows (there is a unique constraint on (from_stop_I,
        # to_stop_I)) and then we update all rows.
        cur.execute(stmt)
        cur2.executemany('INSERT OR IGNORE INTO stop_distances '
                         '(min_transfer_time, from_stop_I, to_stop_I) '
                         'VALUES (?,?,?)',
                         cur)
        # Now, run again to do UPDATE any pre-existing rows.
        cur.execute(stmt)
        cur2.executemany('UPDATE stop_distances '
                         'SET min_transfer_time=? '
                         'WHERE from_stop_I=? and to_stop_I=?',
                         cur)
        conn.commit()

        # Add timed transfers (transfer_type=1).  This is added with
        # timed_transfer=1 and min_transfer_time=0.  Again, first we
        # add missing rows, and then we update the relevant rows.
        stmt = ('SELECT from_stop_I, to_stop_I '
                'FROM transfers '
                'WHERE transfer_type=1 '
                'and from_stop_I!=to_stop_I')
        cur.execute(stmt)
        cur2.executemany('INSERT OR IGNORE INTO stop_distances '
                         '(from_stop_I, to_stop_I) '
                         'VALUES (?,?)',
                         cur)
        cur.execute(stmt)
        cur2.executemany('UPDATE stop_distances '
                         'SET timed_transfer=1, '
                         '    min_transfer_time=0 '
                         'WHERE from_stop_I=? and to_stop_I=?',
                         cur)
        conn.commit()

        # Excluded transfers.  Delete any transfer point with
        # transfer_type=3.
        cur = conn.cursor()
        cur2 = conn.cursor()
        cur.execute('SELECT from_stop_I, to_stop_I '
                    'FROM transfers '
                    'WHERE transfer_type=3')
        cur2.executemany('DELETE FROM stop_distances '
                         'WHERE from_stop_I=? and to_stop_I=?',
                         cur)
        conn.commit()

        # Calculate any `d`s missing because of inserted rows in the
        # previous two steps.
        cur.execute('UPDATE stop_distances '
                    'SET d=CAST (find_distance('
                    ' (SELECT lat FROM stops WHERE stop_I=from_stop_I), '
                    ' (SELECT lon FROM stops WHERE stop_I=from_stop_I), '
                    ' (SELECT lat FROM stops WHERE stop_I=to_stop_I), '
                    ' (SELECT lon FROM stops WHERE stop_I=to_stop_I)  ) '
                    ' AS INT)'
                    'WHERE d ISNULL'
                    )
        conn.commit()

    def export_stop_distances(self, conn, f_out):
        cur = conn.cursor()
        cur.execute('SELECT '
                    'from_stop_I, to_stop_I, '
                    'S1.lat, S1.lon, S2.lat, S2.lon, '
                    'd, '
                    'min_transfer_time '
                    'FROM stop_distances '
                    'LEFT JOIN stops S1 ON (from_stop_I=S1.stop_I)'
                    'LEFT JOIN stops S2 ON (to_stop_I  =S2.stop_I)'
                    )
        f_out.write('#from_stop_I,to_stop_I,'
                    'lat1,lon1,lat2,lon2,'
                    'd,min_transfer_time\n')
        for row in cur:
            f_out.write(','.join(str(x) for x in row) + '\n')



import unittest
from unittest import TestCase

import numpy
from matplotlib import pyplot as plt

from gtfspy.routing.label import LabelTimeSimple
from gtfspy.routing.node_profile_simple import NodeProfileSimple
from gtfspy.routing.node_profile_analyzer_time import NodeProfileAnalyzerTime


class TestNodeProfileAnalyzerTime(TestCase):

    def test_trip_duration_statistics_empty_profile(self):
        profile = NodeProfileSimple()
        analyzer = NodeProfileAnalyzerTime.from_profile(profile, 0, 10)
        self.assertEqual(float('inf'), analyzer.max_trip_duration())
        self.assertEqual(float('inf'), analyzer.min_trip_duration())
        self.assertEqual(float('inf'), analyzer.mean_trip_duration())
        self.assertEqual(float('inf'), analyzer.median_trip_duration())

        self.assertEqual(float('inf'), analyzer.max_temporal_distance())
        self.assertEqual(float('inf'), analyzer.min_temporal_distance())
        self.assertEqual(float('inf'), analyzer.mean_temporal_distance())
        self.assertEqual(float('inf'), analyzer.median_temporal_distance())

    def test_trip_duration_statistics_simple(self):
        pairs = [
            LabelTimeSimple(1.0, 2.0),
            LabelTimeSimple(2.0, 4.0),
            LabelTimeSimple(4.0, 5.0)
        ]
        profile = NodeProfileSimple()
        for pair in pairs:
            profile.update_pareto_optimal_tuples(pair)
        analyzer = NodeProfileAnalyzerTime.from_profile(profile, 0, 100)
        self.assertAlmostEqual(2, analyzer.max_trip_duration())
        self.assertAlmostEqual(1, analyzer.min_trip_duration())
        self.assertAlmostEqual(4 / 3.0, analyzer.mean_trip_duration())
        self.assertAlmostEqual(1, analyzer.median_trip_duration())

    def test_temporal_distance_statistics(self):
        pairs = [
            LabelTimeSimple(1, 2),
            LabelTimeSimple(2, 4),
            LabelTimeSimple(4, 5)
        ]
        profile = NodeProfileSimple()
        for pair in pairs:
            profile.update_pareto_optimal_tuples(pair)

        analyzer = NodeProfileAnalyzerTime.from_profile(profile, 0, 3)
        self.assertAlmostEqual(4 - 1, analyzer.max_temporal_distance())  # 1 -wait-> 2 -travel->4
        self.assertAlmostEqual(1, analyzer.min_temporal_distance())
        self.assertAlmostEqual((1.5 * 1 + 2.5 * 1 + 2.5 * 1) / 3., analyzer.mean_temporal_distance())
        self.assertAlmostEqual(2.25, analyzer.median_temporal_distance())

    def test_temporal_distances_no_transit_trips_within_range(self):
        pairs = [
            LabelTimeSimple(departure_time=11, arrival_time_target=12),
        ]
        profile = NodeProfileSimple(walk_to_target_duration=5)
        for pair in pairs:
            profile.update_pareto_optimal_tuples(pair)
        analyzer = NodeProfileAnalyzerTime.from_profile(profile, 0, 10)
        self.assertAlmostEqual(5, analyzer.max_temporal_distance())
        self.assertAlmostEqual(2, analyzer.min_temporal_distance())
        self.assertAlmostEqual((7 * 5 + 3 * (5 + 2) / 2.) / 10.0, analyzer.mean_temporal_distance())
        self.assertAlmostEqual(5, analyzer.median_temporal_distance())

    def test_temporal_distances_no_transit_trips_within_range_and_no_walk(self):
        pairs = [
            LabelTimeSimple(departure_time=11, arrival_time_target=12),
        ]
        profile = NodeProfileSimple(walk_to_target_duration=float('inf'))
        for pair in pairs:
            profile.update_pareto_optimal_tuples(pair)
        analyzer = NodeProfileAnalyzerTime.from_profile(profile, 0, 10)
        self.assertAlmostEqual(12, analyzer.max_temporal_distance())
        self.assertAlmostEqual(2, analyzer.min_temporal_distance())
        self.assertAlmostEqual((12 + 2) / 2.0, analyzer.mean_temporal_distance())
        self.assertAlmostEqual((12 + 2) / 2.0, analyzer.median_temporal_distance())

    def test_time_offset(self):
        max_distances = []
        for offset in [0, 10, 100, 1000]:
            labels = [
                LabelTimeSimple(departure_time=7248 + offset, arrival_time_target=14160 + offset),
            ]
            profile = NodeProfileSimple(walk_to_target_duration=float('inf'))
            for label in labels:
                profile.update_pareto_optimal_tuples(label)
            analyzer = NodeProfileAnalyzerTime.from_profile(profile, 0 + offset, 7200 + offset)
            max_distances.append(analyzer.max_temporal_distance())
        max_distances = numpy.array(max_distances)
        assert((max_distances == max_distances[0]).all())
        # self.assertAlmostEqual(12, analyzer.max_temporal_distance())
        # self.assertAlmostEqual(2, analyzer.min_temporal_distance())
        # self.assertAlmostEqual((12 + 2) / 2.0, analyzer.mean_temporal_distance())
        # self.assertAlmostEqual((12 + 2) / 2.0, analyzer.median_temporal_distance())

    def test_temporal_distance_statistics_with_walk(self):
        pt1 = LabelTimeSimple(departure_time=1, arrival_time_target=2)
        pt2 = LabelTimeSimple(departure_time=4, arrival_time_target=5)   # not taken into account by the analyzer
        profile = NodeProfileSimple(1.5)
        assert isinstance(pt1, LabelTimeSimple), type(pt1)
        profile.update_pareto_optimal_tuples(pt1)
        profile.update_pareto_optimal_tuples(pt2)
        analyzer = NodeProfileAnalyzerTime.from_profile(profile, 0, 3)
        self.assertAlmostEqual(1.5, analyzer.max_temporal_distance())  # 1 -wait-> 2 -travel->4
        self.assertAlmostEqual(1, analyzer.min_temporal_distance())
        self.assertAlmostEqual((2.5 * 1.5 + 0.5 * 1.25) / 3., analyzer.mean_temporal_distance())
        self.assertAlmostEqual(1.5, analyzer.median_temporal_distance())

    def test_temporal_distance_statistics_with_walk2(self):
        pt1 = LabelTimeSimple(departure_time=10, arrival_time_target=30)
        profile = NodeProfileSimple(25)
        profile.update_pareto_optimal_tuples(pt1)
        analyzer = NodeProfileAnalyzerTime.from_profile(profile, 0, 10)
        # analyzer.plot_temporal_distance_profile()
        # plt.show()

        self.assertAlmostEqual(25, analyzer.max_temporal_distance())  # 1 -wait-> 2 -travel->4
        self.assertAlmostEqual(20, analyzer.min_temporal_distance())
        self.assertAlmostEqual((7.5 * 25 + 2.5 * 20) / 10.0, analyzer.mean_temporal_distance())
        self.assertAlmostEqual(25, analyzer.median_temporal_distance())

    def test_temporal_distance_pdf_with_walk(self):
        profile = NodeProfileSimple(25)
        pt1 = LabelTimeSimple(10, 30)
        profile.update_pareto_optimal_tuples(pt1)
        analyzer = NodeProfileAnalyzerTime.from_profile(profile, 0, 10)

        self.assertEqual(len(analyzer.profile_block_analyzer._temporal_distance_pdf()), 3)

        split_points, densities, delta_peaks = analyzer.profile_block_analyzer._temporal_distance_pdf()
        self.assertEqual(len(split_points), 2)
        self.assertEqual(split_points[0], 20)
        self.assertEqual(split_points[1], 25)

        self.assertEqual(len(densities), 1)
        self.assertEqual(densities[0], 0.1)

        self.assertIn(25, delta_peaks)
        self.assertEqual(delta_peaks[25], 0.5)


    @unittest.skip("Skipping plotting test")
    def test_all_plots(self):
        profile = NodeProfileSimple(25)
        pt1 = LabelTimeSimple(departure_time=10, arrival_time_target=30)
        profile.update_pareto_optimal_tuples(pt1)
        analyzer = NodeProfileAnalyzerTime.from_profile(profile, 0, 10)
        analyzer.plot_temporal_distance_profile(plot_tdist_stats=True)
        analyzer.plot_temporal_distance_cdf()
        analyzer.plot_temporal_distance_pdf()
        plt.show()

        profile = NodeProfileSimple()
        profile.update_pareto_optimal_tuples(LabelTimeSimple(departure_time=2 * 60, arrival_time_target=11 * 60))
        profile.update_pareto_optimal_tuples(LabelTimeSimple(departure_time=20 * 60, arrival_time_target=25 * 60))
        profile.update_pareto_optimal_tuples(LabelTimeSimple(departure_time=40 * 60, arrival_time_target=45 * 60))
        analyzer = NodeProfileAnalyzerTime.from_profile(profile, 0, 60 * 60)
        analyzer.plot_temporal_distance_profile()
        analyzer.plot_temporal_distance_cdf()
        analyzer.plot_temporal_distance_pdf()

        profile = NodeProfileSimple()
        profile.update_pareto_optimal_tuples(LabelTimeSimple(departure_time=2 * 60, arrival_time_target=3 * 60))
        profile.update_pareto_optimal_tuples(LabelTimeSimple(departure_time=4 * 60, arrival_time_target=25 * 60))
        analyzer = NodeProfileAnalyzerTime.from_profile(profile, 0, 5 * 60)
        analyzer.plot_temporal_distance_profile()
        analyzer.plot_temporal_distance_cdf()
        analyzer.plot_temporal_distance_pdf()

        pt1 = LabelTimeSimple(departure_time=1, arrival_time_target=2)
        pt2 = LabelTimeSimple(departure_time=4, arrival_time_target=5)   # not taken into account by the analyzer
        profile = NodeProfileSimple(1.5)
        profile.update_pareto_optimal_tuples(pt1)
        profile.update_pareto_optimal_tuples(pt2)
        analyzer = NodeProfileAnalyzerTime.from_profile(profile, 0, 3)
        analyzer.plot_temporal_distance_profile()
        analyzer.plot_temporal_distance_cdf()

        plt.show()




from __future__ import print_function

import warnings
from collections import defaultdict

import datetime
import matplotlib
import numpy
import matplotlib.pyplot as plt
import pytz
from matplotlib import lines

from matplotlib import dates as md
from matplotlib.colors import ListedColormap, LinearSegmentedColormap

from gtfspy.routing.fastest_path_analyzer import FastestPathAnalyzer
from gtfspy.routing.node_profile_multiobjective import NodeProfileMultiObjective
from gtfspy.routing.label import LabelTimeWithBoardingsCount, compute_pareto_front, LabelTimeSimple
from gtfspy.routing.node_profile_analyzer_time import NodeProfileAnalyzerTime
from gtfspy.routing.node_profile_simple import NodeProfileSimple
from gtfspy.routing.profile_block_analyzer import ProfileBlockAnalyzer


def _check_for_no_labels_for_n_veh_counts(func):
    def wrapper(self):
        assert (isinstance(self, NodeProfileAnalyzerTimeAndVehLegs))
        if len(self._labels_within_time_frame) == 0:
            if self._walk_to_target_duration is None:
                return 0
            else:
                return float('nan')
        else:
            return func(self)

    return wrapper


def _if_no_labels_return_inf(func):
    def wrapper(self):
        if self._labels_within_time_frame:
            return func(self)
        else:
            return float('inf')
    return wrapper


def _truncate_colormap(cmap, minval=0.0, maxval=1.0, n=100):
    """
    Truncates a colormap to use.
    Code originall from http://stackoverflow.com/questions/18926031/how-to-extract-a-subset-of-a-colormap-as-a-new-colormap-in-matplotlib
    """
    new_cmap = LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=minval, b=maxval),
        cmap(numpy.linspace(minval, maxval, n))
    )
    return new_cmap


class NodeProfileAnalyzerTimeAndVehLegs:

    @classmethod
    def from_profile(cls, node_profile, start_time_dep, end_time_dep):
        assert (node_profile.label_class == LabelTimeWithBoardingsCount)
        return NodeProfileAnalyzerTimeAndVehLegs(node_profile.get_final_optimal_labels(),
                                                 node_profile.get_walk_to_target_duration(),
                                                 start_time_dep,
                                                 end_time_dep)

    def __init__(self, labels, walk_to_target_duration, start_time_dep, end_time_dep):
        """
        Initialize the data structures required by

        Parameters
        ----------
        node_profile: NodeProfileMultiObjective
        """
        self._node_profile_final_labels = labels
        self.start_time_dep = start_time_dep
        self.end_time_dep = end_time_dep
        self.all_labels = [label for label in self._node_profile_final_labels if
                           (start_time_dep <= label.departure_time <= end_time_dep)]
        after_label_candidates = [label for label in self._node_profile_final_labels if
                                  (label.departure_time > self.end_time_dep)]
        after_label_candidates.sort(key=lambda el: (el.arrival_time_target, el.n_boardings))
        min_n_boardings_observed = float('inf')
        after_labels = []
        for candidate_after_label in after_label_candidates:
            if candidate_after_label.n_boardings < min_n_boardings_observed:
                after_labels.append(candidate_after_label)
                min_n_boardings_observed = candidate_after_label.n_boardings

        self.all_labels.extend(after_labels)
        if len(after_labels) is 0:
            self._labels_within_time_frame = self.all_labels
        else:
            self._labels_within_time_frame = self.all_labels[:-len(after_labels)]

        self._walk_to_target_duration = walk_to_target_duration
        self._n_boardings_to_simple_time_analyzers = {}
        self._transfers_on_fastest_paths_analyzer = self._get_transfers_on_fastest_path_analyzer()

    def __get_fastest_path_analyzer(self):
        return FastestPathAnalyzer(self.all_labels,
                                   self.start_time_dep,
                                   self.end_time_dep,
                                   walk_duration=self._walk_to_target_duration,
                                   label_props_to_consider=["n_boardings"])

    def _get_transfers_on_fastest_path_analyzer(self):
        fp_analyzer = self.__get_fastest_path_analyzer()
        if self._walk_to_target_duration < float('inf'):
            cutoff_value = 0
        else:
            cutoff_value = float('inf')
        return fp_analyzer.get_prop_analyzer_flat("n_boardings", float('inf'), cutoff_value)

    def min_n_boardings(self):
        if self._walk_to_target_duration < float('inf'):
            return 0
        else:
            if len(self.all_labels) is 0:
                return float('inf')
            else:
                return min([label.n_boardings for label in self.all_labels])

    def min_n_boardings_on_shortest_paths(self):
        return self._transfers_on_fastest_paths_analyzer.min()

    def max_n_boardings_on_shortest_paths(self):
        return self._transfers_on_fastest_paths_analyzer.max()

    def max_finite_n_boardings_on_fastest_paths(self):
        return self._transfers_on_fastest_paths_analyzer.largest_finite_distance()

    def mean_n_boardings_on_shortest_paths(self):
        import math
        mean = self._transfers_on_fastest_paths_analyzer.mean()
        if math.isnan(mean):
            mean = self._transfers_on_fastest_paths_analyzer.mean()
        return mean

    def median_n_boardings_on_shortest_paths(self):
        return self._transfers_on_fastest_paths_analyzer.median()

    def get_time_profile_analyzer(self, max_n_boardings=None):
        """
        Parameters
        ----------
        max_n_boardings: int
            The maximum number of boardings allowed for the labels used to construct the "temporal distance profile"

        Returns
        -------
        analyzer: NodeProfileAnalyzerTime
        """
        if max_n_boardings is None:
            max_n_boardings = self.max_trip_n_boardings()
        # compute only if not yet computed
        if not max_n_boardings in self._n_boardings_to_simple_time_analyzers:
            if max_n_boardings == 0:
                valids = []
            else:
                candidate_labels = [LabelTimeSimple(label.departure_time, label.arrival_time_target)
                                    for label in self._node_profile_final_labels if
                                    ((self.start_time_dep <= label.departure_time)
                                     and label.n_boardings <= max_n_boardings)]
                valids = compute_pareto_front(candidate_labels)
            valids.sort(key=lambda label: -label.departure_time)
            profile = NodeProfileSimple(self._walk_to_target_duration)
            for valid in valids:
                profile.update_pareto_optimal_tuples(valid)
            npat = NodeProfileAnalyzerTime.from_profile(profile, self.start_time_dep, self.end_time_dep)
            self._n_boardings_to_simple_time_analyzers[max_n_boardings] = npat
        return self._n_boardings_to_simple_time_analyzers[max_n_boardings]

    @_check_for_no_labels_for_n_veh_counts
    def max_trip_n_boardings(self):
        return numpy.max([label.n_boardings for label in self._labels_within_time_frame])

    @_check_for_no_labels_for_n_veh_counts
    def min_trip_n_boardings(self):
        values = [label.n_boardings for label in self._labels_within_time_frame]
        min_val = numpy.min(values)
        return min_val

    @_check_for_no_labels_for_n_veh_counts
    def mean_trip_n_boardings(self):
        return numpy.mean([label.n_boardings for label in self._labels_within_time_frame])

    @_check_for_no_labels_for_n_veh_counts
    def median_trip_n_boardings(self):
        return numpy.median([label.n_boardings for label in self._labels_within_time_frame])
 
    @_check_for_no_labels_for_n_veh_counts
    def n_boardings_on_fastest_trip(self):
        return min(self._labels_within_time_frame, key=lambda label: label.arrival_time_target - label.departure_time).n_boardings 

    @_if_no_labels_return_inf
    def min_temporal_distance(self):
        result = self.get_time_profile_analyzer().min_temporal_distance()
        assert (result >= 0), result
        return result

    @_if_no_labels_return_inf
    def max_temporal_distance(self):
        return self.get_time_profile_analyzer().max_temporal_distance()

    @_if_no_labels_return_inf
    def median_temporal_distance(self):
        return self.get_time_profile_analyzer().median_temporal_distance()

    @_if_no_labels_return_inf
    def mean_temporal_distance(self):
        return self.get_time_profile_analyzer().mean_temporal_distance()

    @_if_no_labels_return_inf
    def min_trip_duration(self):
        return self.get_time_profile_analyzer().min_trip_duration()

    @_if_no_labels_return_inf
    def max_trip_duration(self):
        return self.get_time_profile_analyzer().max_trip_duration()

    @_if_no_labels_return_inf
    def median_trip_duration(self):
        return self.get_time_profile_analyzer().median_trip_duration()

    @_if_no_labels_return_inf
    def mean_trip_duration(self):
        return self.get_time_profile_analyzer().mean_trip_duration()

    def mean_temporal_distance_with_min_n_boardings(self):
        min_n_boardings = self.min_n_boardings()
        min_n_boardings_analyzer = self.get_time_profile_analyzer(min_n_boardings)
        return min_n_boardings_analyzer.mean_temporal_distance()

    def min_temporal_distance_with_min_n_boardings(self):
        min_n_boardings = self.min_n_boardings()
        min_n_boardings_analyzer = self.get_time_profile_analyzer(min_n_boardings)
        return min_n_boardings_analyzer.min_temporal_distance()

    def median_temporal_distances(self, min_n_boardings=None, max_n_boardings=None):
        """
        Returns
        -------
        mean_temporal_distances: list
            list indices encode the number of vehicle legs each element
            in the list tells gets the mean temporal distance
        """
        if min_n_boardings is None:
            min_n_boardings = 0

        if max_n_boardings is None:
            max_n_boardings = self.max_trip_n_boardings()
            if max_n_boardings is None:
                max_n_boardings = 0

        median_temporal_distances = [float('inf') for _ in range(min_n_boardings, max_n_boardings + 1)]
        for n_boardings in range(min_n_boardings, max_n_boardings + 1):
            simple_analyzer = self.get_time_profile_analyzer(n_boardings)
            median_temporal_distances[n_boardings] = simple_analyzer.median_temporal_distance()
        return median_temporal_distances

    @classmethod
    def _get_colors_for_boardings(cls, min_n_boardings, max_n_boardings):
        cmap = NodeProfileAnalyzerTimeAndVehLegs.get_colormap_for_boardings(max_n_boardings)
        colors = [cmap(float(n_boardings) / max_n_boardings) for n_boardings in range(int(max_n_boardings) + 1)]
        return colors[min_n_boardings:max_n_boardings + 1]

    @classmethod
    def get_colormap_for_boardings(cls, max_n_boardings=None):
        n_default = 5
        if max_n_boardings in [float('nan'), None]:
            max_n_boardings = n_default
        from matplotlib import cm
        cmap = cm.get_cmap("cubehelix_r")
        start = 0.1
        end = 0.9
        if max_n_boardings is 0:
            step = 0
        else:
            divider = max(n_default, max_n_boardings)
            step = (end - start) / divider
        truncated = _truncate_colormap(cmap, start, start + step * max_n_boardings)
        return truncated

    @classmethod
    def _multiply_color_saturation(cls, color, multiplier):
        hsv = matplotlib.colors.rgb_to_hsv(color[:3])
        rgb = matplotlib.colors.hsv_to_rgb((hsv[0], hsv[1] * multiplier, hsv[2]))
        return list(iter(rgb)) + [1]

    @classmethod
    def _multiply_color_brightness(cls, color, multiplier):
        hsv = matplotlib.colors.rgb_to_hsv(color[:3])
        rgb = matplotlib.colors.hsv_to_rgb((hsv[0], hsv[1], max(0, min(1, hsv[2] * multiplier))))
        return list(iter(rgb)) + [1]

    def _get_fill_and_line_colors(self, min_n, max_n):
        colors = self._get_colors_for_boardings(min_n, max_n)
        n_boardings_range = range(min_n, max_n + 1)
        nboardings_to_color = {n: colors[i] for i, n in enumerate(n_boardings_range)}

        n_boardings_to_line_color = {}
        n_boardings_to_fill_color = {}

        #
        rgbs = [color_tuple[:3] for color_tuple in nboardings_to_color.values()]
        hsvs = matplotlib.colors.rgb_to_hsv(rgbs)
        max_saturation = max([hsv[1] for hsv in hsvs])
        line_saturation_multiplier = 1 / max_saturation

        for n, color_tuple in nboardings_to_color.items():
            c = NodeProfileAnalyzerTimeAndVehLegs._multiply_color_saturation(color_tuple, line_saturation_multiplier)
            c = NodeProfileAnalyzerTimeAndVehLegs._multiply_color_brightness(c, 1)
            n_boardings_to_line_color[n] = c

            c = NodeProfileAnalyzerTimeAndVehLegs._multiply_color_brightness(color_tuple, 1.2)
            c = NodeProfileAnalyzerTimeAndVehLegs._multiply_color_saturation(c, 0.8)
            n_boardings_to_fill_color[n] = c
        return n_boardings_to_fill_color, n_boardings_to_line_color

    @classmethod
    def n_boardings_to_label(self, n):
        if n is 0:
            return "walk"
        elif n is 1:
            return "1 boarding"
        else:
            return str(n) + " boardings"

    def plot_new_transfer_temporal_distance_profile(self,
                                                    timezone=None,
                                                    format_string="%Y-%m-%d %H:%M:%S",
                                                    duration_divider=60.0,
                                                    ax=None,
                                                    plot_journeys=False,
                                                    highlight_fastest_path=True,
                                                    default_lw=5,
                                                    ncol_legend=1,
                                                    fastest_path_lw=3,
                                                    legend_alpha=0.9,
                                                    journey_letters=None,
                                                    legend_font_size=None):
        max_n = self.max_finite_n_boardings_on_fastest_paths()
        min_n = self.min_n_boardings()
        if self._walk_to_target_duration < float('inf'):
            min_n = 0
        if max_n is None:
            return None
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        fig = ax.figure
        assert (isinstance(ax, matplotlib.axes.Axes))

        if timezone is None:
            warnings.warn("Warning: No timezone specified, defaulting to UTC")
            timezone = pytz.timezone("Etc/UTC")

        def _ut_to_unloc_datetime(ut):
            dt = datetime.datetime.fromtimestamp(ut, timezone)
            return dt.replace(tzinfo=None)

        if format_string:
            x_axis_formatter = md.DateFormatter(format_string)
            ax.xaxis.set_major_formatter(x_axis_formatter)
        else:
            _ut_to_unloc_datetime = lambda x: x

        ax.set_xlim(
            _ut_to_unloc_datetime(self.start_time_dep),
            _ut_to_unloc_datetime(self.end_time_dep)
        )

        n_boardings_range = range(min_n, max_n + 1)
        n_boardings_to_lw = {n: default_lw for i, n in enumerate(n_boardings_range)}

        n_boardings_to_fill_color, n_boardings_to_line_color = self._get_fill_and_line_colors(min_n, max_n)

        # get all trips ordered by departure time


        deptime_ordered_labels = sorted(list(self.all_labels), key=lambda x: x.departure_time)

        n_boardings_to_labels = defaultdict(list)
        for journey_label in deptime_ordered_labels:
            n_boardings_to_labels[journey_label.n_boardings].append(journey_label)

        walk_duration = self._walk_to_target_duration / duration_divider
        if walk_duration < float('inf'):
            xs = [_ut_to_unloc_datetime(x) for x in [self.start_time_dep, self.end_time_dep]]
            ax.plot(xs, [walk_duration, walk_duration], lw=n_boardings_to_lw[0], color=n_boardings_to_line_color[0])
            ax.fill_between(xs, 0, walk_duration, color=n_boardings_to_fill_color[0])
            max_tdist = walk_duration
        else:
            max_tdist = 0

        journeys = []
        for n_boardings in n_boardings_range:
            if n_boardings == 0:
                # dealt above
                continue
            prev_analyzer = self.get_time_profile_analyzer(n_boardings - 1)
            prev_profile_block_analyzer = prev_analyzer.profile_block_analyzer
            assert (isinstance(prev_profile_block_analyzer, ProfileBlockAnalyzer))
            labels = n_boardings_to_labels[n_boardings]

            prev_was_larger_already = False
            for i, journey_label in enumerate(labels):
                if journey_label.departure_time > self.end_time_dep:
                    if prev_was_larger_already:
                        continue

                prev_dep_time = self.start_time_dep
                if i is not 0:
                    prev_dep_time = labels[i - 1].departure_time
                # this could perhaps be made a while loop of some sort
                # to not loop over things multiple times
                for block in prev_profile_block_analyzer._profile_blocks:
                    if block.distance_end != block.distance_start:
                        if block.distance_end <= journey_label.duration() + (
                                    journey_label.departure_time - block.end_time):
                            prev_dep_time = max(prev_dep_time, block.end_time)
                    elif block.distance_end == block.distance_start:
                        # look for the time when
                        waiting_time = (block.distance_end - journey_label.duration())
                        prev_dep_time = max(prev_dep_time, journey_label.departure_time - waiting_time)
                # prev dep time is now known
                waiting_time = journey_label.departure_time - prev_dep_time
                lw = n_boardings_to_lw[n_boardings]
                xs = [_ut_to_unloc_datetime(prev_dep_time), _ut_to_unloc_datetime(journey_label.departure_time)]
                ys = numpy.array([journey_label.duration() + waiting_time, journey_label.duration()]) / duration_divider
                max_tdist = max(ys[0], max_tdist)
                ax.plot(xs, ys,
                        color=n_boardings_to_line_color[n_boardings],  # "k",
                        lw=lw)
                ax.fill_between(xs, 0, ys, color=n_boardings_to_fill_color[n_boardings])
                if plot_journeys:
                    journeys.append((xs[1], ys[1]))

        legend_patches = []
        for n_boardings in n_boardings_range:
            text = self.n_boardings_to_label(n_boardings)
            p = lines.Line2D([0, 1], [0, 1], lw=n_boardings_to_lw[n_boardings],
                             color=n_boardings_to_line_color[n_boardings],
                             label=text)
            legend_patches.append(p)

        if highlight_fastest_path:
            fastest_path_time_analyzer = self.get_time_profile_analyzer()
            vlines, slopes = fastest_path_time_analyzer.profile_block_analyzer.get_vlines_and_slopes_for_plotting()
            lw = fastest_path_lw
            ls = "--"
            for vline in vlines:
                ax.plot([_ut_to_unloc_datetime(x) for x in vline['x']], numpy.array(vline['y']) / duration_divider,
                        ls=":", lw=lw / 2, color="k")
            for slope in slopes:
                ax.plot([_ut_to_unloc_datetime(x) for x in slope['x']], numpy.array(slope['y']) / duration_divider,
                        ls=ls, color="k", lw=lw)
            p = lines.Line2D([0, 1], [0, 1], ls=ls, lw=lw, color="k", label="fastest path profile")
            legend_patches.append(p)

        if plot_journeys:
            if journey_letters is None:
                journey_letters = iter("ABCDEFGHIJKLMN")
            journeys.sort()
            for (x, y), letter in zip(journeys, journey_letters):
                if x < _ut_to_unloc_datetime(self.end_time_dep):
                    ax.plot(x, y, "o", ms=8, color="k")
                    ax.text(x + datetime.timedelta(seconds=(self.end_time_dep - self.start_time_dep) / 40.),
                            y, letter, va="center", ha="left")
            p = lines.Line2D([0, 0], [1, 1], ls="", marker="o", ms=8, color="k", label="journeys")
            legend_patches.append(p)

        if legend_font_size is None:
            legend_font_size = 12
        leg = ax.legend(handles=legend_patches, loc="best", ncol=ncol_legend, fancybox=True, prop={"size": legend_font_size})
        leg.get_frame().set_alpha(legend_alpha)
        ax.set_ylim(0, 1.1 * max_tdist)
        ax.set_xlabel("Departure time $t_\\mathrm{dep}$")
        ax.set_ylabel("Temporal distance $\\tau$")
        return fig

    def plot_temporal_distance_pdf_horizontal(self, use_minutes=True, ax=None, duration_divider=60.0,
                                              legend_font_size=None, legend_loc=None):
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)

        blocks = self.__get_fastest_path_analyzer().get_fastest_path_temporal_distance_blocks()

        walking_is_fastest_time = 0
        non_walk_blocks = []
        tdist_split_points = set()
        for block in blocks:
            if block.is_flat(): # test for walk
                walking_is_fastest_time += block.width()
            else:
                non_walk_blocks.append(block)
                tdist_split_points.add(block.distance_end)
                tdist_split_points.add(block.distance_start)

        distance_split_points_ordered = numpy.array(sorted(list(tdist_split_points)))
        temporal_distance_split_widths = distance_split_points_ordered[1:] - distance_split_points_ordered[:-1]

        non_walk_blocks_total_time = sum((block.width() for block in non_walk_blocks))
        assert (
            numpy.isclose(walking_is_fastest_time + non_walk_blocks_total_time,
                          self.end_time_dep - self.start_time_dep)
        )

        min_n_boardings = int(self.min_n_boardings())
        max_n_boardings = int(self.max_finite_n_boardings_on_fastest_paths())
        fill_colors, line_colors = self._get_fill_and_line_colors(min_n_boardings, max_n_boardings)


        temporal_distance_values_to_plot = []
        for x in distance_split_points_ordered:
            temporal_distance_values_to_plot.append(x)
            temporal_distance_values_to_plot.append(x)
        temporal_distance_values_to_plot = numpy.array(temporal_distance_values_to_plot)

        pdf_values_to_plot_by_n_boardings = {}
        pdf_areas = {}

        for n_boardings in range(min_n_boardings, max_n_boardings + 1):
            # if walking, the block has no "n_boardings" attribute
            blocks_now = [block for block in non_walk_blocks if block["n_boardings"] >= n_boardings]

            journey_counts = numpy.zeros(len(temporal_distance_split_widths))
            for block_now in blocks_now:
                first_index = numpy.searchsorted(distance_split_points_ordered, block_now.distance_end)
                last_index = numpy.searchsorted(distance_split_points_ordered, block_now.distance_start)
                journey_counts[first_index:last_index] += 1

            part_pdf = journey_counts / (self.end_time_dep - self.start_time_dep)
            pdf_areas[n_boardings] = sum(part_pdf * temporal_distance_split_widths)

            pdf_values_to_plot = [0]
            for pdf_value in part_pdf:
                pdf_values_to_plot.append(pdf_value)
                pdf_values_to_plot.append(pdf_value)
            pdf_values_to_plot.append(0)
            pdf_values_to_plot_by_n_boardings[n_boardings] = numpy.array(pdf_values_to_plot)

        if walking_is_fastest_time > 0:
            text = "$P(\\mathrm{walk}) = " + ("%.2f$" % (walking_is_fastest_time/(self.end_time_dep - self.start_time_dep)))
            ax.plot([0, 10], [self._walk_to_target_duration / duration_divider, self._walk_to_target_duration / duration_divider], color=line_colors[0],
                    lw=5, label=text, zorder=10)

        for n_boardings in range(max(1, min_n_boardings), max_n_boardings + 1):
            if n_boardings is max_n_boardings:
                prob = pdf_areas[n_boardings]
            else:
                prob = pdf_areas[n_boardings] - pdf_areas[n_boardings+1]
            pdf_values_to_plot = pdf_values_to_plot_by_n_boardings[n_boardings]
            if n_boardings is 0:
                label = "$P(\\mathrm{walk})= %.2f $" % (prob)
            else:
                label = "$P(b=" + str(n_boardings) + ")= %.2f $" % (prob)
                # "$P_\\mathrm{" + self.n_boardings_to_label(n_boardings).replace(" ", "\\,") + "} = %.2f $" % (prob)
            ax.fill_betweenx(temporal_distance_values_to_plot / duration_divider,
                             pdf_values_to_plot * duration_divider,
                             label=label,
                             color=fill_colors[n_boardings], zorder=n_boardings)
            ax.plot(pdf_values_to_plot * duration_divider, temporal_distance_values_to_plot / duration_divider,
                    color=line_colors[n_boardings], zorder=n_boardings)
            # get counts for each plot

        if legend_font_size is None:
            legend_font_size = 12
        if legend_loc is None:
            legend_loc = "best"
        leg = ax.legend(loc=legend_loc, fancybox=True, prop={"size": legend_font_size})
        leg.get_frame().set_alpha(0.9)
        return ax

    def plot_fastest_temporal_distance_profile(self, timezone=None, **kwargs):
        max_n = self.max_trip_n_boardings()
        if "ax" not in kwargs:
            fig = plt.figure(figsize=(10, 6))
            ax = fig.add_subplot(111)
            kwargs["ax"] = ax
        npat = self.get_time_profile_analyzer(max_n)
        fig = npat.plot_temporal_distance_profile(timezone=timezone,
                                                  **kwargs)
        return fig

    def n_pareto_optimal_trips(self):
        """
        Get number of pareto-optimal trips

        Returns
        -------
        n_trips: float
        """
        return float(len(self._labels_within_time_frame))

    @staticmethod
    def all_measures_and_names_as_lists():
        NPA = NodeProfileAnalyzerTimeAndVehLegs
        profile_summary_methods = [
            NPA.max_trip_duration,
            NPA.mean_trip_duration,
            NPA.median_trip_duration,
            NPA.min_trip_duration,
            NPA.max_temporal_distance,
            NPA.mean_temporal_distance,
            NPA.median_temporal_distance,
            NPA.min_temporal_distance,
            NPA.n_pareto_optimal_trips,
            NPA.min_n_boardings,
            NPA.min_trip_n_boardings,
            NPA.max_trip_n_boardings,
            NPA.mean_trip_n_boardings,
            NPA.median_trip_n_boardings,
            NPA.mean_n_boardings_on_shortest_paths,
            NPA.min_n_boardings_on_shortest_paths,
            NPA.max_n_boardings_on_shortest_paths,
            NPA.median_n_boardings_on_shortest_paths,
            NPA.mean_temporal_distance_with_min_n_boardings,
            NPA.min_temporal_distance_with_min_n_boardings
        ]
        profile_observable_names = [
            "max_trip_duration",
            "mean_trip_duration",
            "median_trip_duration",
            "min_trip_duration",
            "max_temporal_distance",
            "mean_temporal_distance",
            "median_temporal_distance",
            "min_temporal_distance",
            "n_pareto_optimal_trips",
            "min_n_boardings",
            "min_trip_n_boardings",
            "max_trip_n_boardings",
            "mean_trip_n_boardings",
            "median_trip_n_boardings",
            "mean_n_boardings_on_shortest_paths",
            "min_n_boardings_on_shortest_paths",
            "max_n_boardings_on_shortest_paths",
            "median_n_boardings_on_shortest_paths",
            "mean_temporal_distance_with_min_n_boardings",
            "min_temporal_distance_with_min_n_boardings"
        ]
        assert (len(profile_summary_methods) == len(profile_observable_names))
        return profile_summary_methods, profile_observable_names

    def get_node_profile_measures_as_dict(self):
        profile_summary_methods, profile_observable_names = self.all_measures_and_names_as_lists()
        profile_measure_dict = {key: value(self) for key, value in zip(profile_observable_names, profile_summary_methods)}
        return profile_measure_dict




from gtfspy.import_loaders.table_loader import TableLoader, decode_six


class RouteLoader(TableLoader):
    fname = 'routes.txt'
    table = 'routes'
    tabledef = '(route_I INTEGER PRIMARY KEY, ' \
               'route_id TEXT UNIQUE NOT NULL, ' \
               'agency_I INT, ' \
               'name TEXT, ' \
               'long_name TEXT, ' \
               'desc TEXT, ' \
               'type INT, ' \
               'url TEXT, ' \
               'color TEXT, ' \
               'text_color TEXT' \
               ')'
    extra_keys = ['agency_I', ]
    extra_values = ['(SELECT agency_I FROM agencies WHERE agency_id=:_agency_id )',
                    ]

    # route_id,agency_id,route_short_name,route_long_name,route_desc,route_type,route_url
    # 1001,HSL,1,Kauppatori - Kapyla,0,http://aikataulut.hsl.fi/linjat/fi/h1_1a.html
    def gen_rows(self, readers, prefixes):
        from gtfspy import extended_route_types
        for reader, prefix in zip(readers, prefixes):
            for row in reader:
                #print (row)
                yield dict(
                    route_id      = prefix + decode_six(row['route_id']),
                    _agency_id    = prefix + decode_six(row['agency_id']) if 'agency_id' in row else None,
                    name          = decode_six(row['route_short_name']),
                    long_name     = decode_six(row['route_long_name']),
                    desc          = decode_six(row['route_desc']) if 'route_desc' in row else None,
                    type          = extended_route_types.ROUTE_TYPE_CONVERSION[int(row['route_type'])],
                    url           = decode_six(row['route_url']) if 'route_url' in row else None,
                    color         = decode_six(row['route_color']) if 'route_color' in row else None,
                    text_color    = decode_six(row['route_text_color']) if 'route_text_color' in row else None,
                )

    @classmethod
    def index(cls, cur):
        # cur.execute('CREATE INDEX IF NOT EXISTS idx_rid ON route (route_id)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_route_name ON routes (name)')


import unittest
import os
import pandas as pd

from gtfspy import util


class TestUtil(unittest.TestCase):

    @staticmethod
    def _approximately_equal(a, b):
        return abs(a - b) / float(abs(a + b)) < 1e-2

    def test_ut_to_utc_datetime_str(self):
        time_ut = 1438300800 + 24299
        assert util.ut_to_utc_datetime_str(time_ut) == "Jul 31 2015 06:44:59"

    def test_wgs84_width(self):
        lat = 60.4192161560059
        lon = 25.3302955627441
        # Test going 100m east
        width = util.wgs84_width(100, lat)
        lon2 = lon + width
        d = util.wgs84_distance(lat, lon, lat, lon2)
        self.assertTrue(self._approximately_equal(d, 100))
        # Test going 100m south
        height = util.wgs84_height(100)
        lat2 = lat + height
        d = util.wgs84_distance(lat, lon, lat2, lon)
        self.assertTrue(self._approximately_equal(d, 100))

    def test_day_seconds_to_str_time(self):
        str_time = util.day_seconds_to_str_time(25 * 3600 + 59 * 60 + 10)
        self.assertTrue(str_time == "25:59:10", "the times can also go over 24 hours")
        str_time = util.day_seconds_to_str_time(0)
        self.assertTrue(str_time == "00:00:00", "day should start at 00:00")

    def test_txt_to_pandas(self):
        source_dir = os.path.join(os.path.dirname(__file__), "test_data")
        txtnames = ['agency', 'routes', 'trips', 'calendar', 'calendar_dates', 'stop_times', 'stops', 'shapes']
        df = util.source_csv_to_pandas(source_dir, txtnames[3])
        self.assertIsInstance(df, pd.DataFrame)
        source_zip = os.path.join(os.path.dirname(__file__), "test_data/test_gtfs.zip")
        df = util.source_csv_to_pandas(source_zip, txtnames[4])
        self.assertIsInstance(df, pd.DataFrame)

    def test_difference_of_pandas_dfs(self):
        dict1 = {"lat": [1, 2, 3, 4, 5], "lon": [5, 6, 7, 8, 9], "data": [123, 342, 345, 123, 543]}
        dict2 = {"lat": [6, 7, 3, 4, 5], "lon": [5, 6, 7, 8, 9], "data": [656, 12, 34, 1112, 43]}
        df1 = pd.DataFrame(dict1)
        df2 = pd.DataFrame(dict2)
        df = util.difference_of_pandas_dfs(df1, df2, ["lat", "lon"])
        self.assertEqual(len(df.index), 2)


from pyximport import install
install()

from unittest import TestCase

from gtfspy.routing.label import LabelTimeSimple, LabelTimeWithBoardingsCount
from gtfspy.routing.node_profile_simple import NodeProfileSimple


class TestNodeProfileSimple(TestCase):

    def test_earliest_arrival_time(self):
        node_profile = NodeProfileSimple()
        self.assertEquals(float("inf"), node_profile.evaluate_earliest_arrival_time_at_target(0, 0))

        node_profile.update_pareto_optimal_tuples(LabelTimeSimple(departure_time=1, arrival_time_target=1))
        self.assertEquals(1, node_profile.evaluate_earliest_arrival_time_at_target(0, 0))

        node_profile.update_pareto_optimal_tuples(LabelTimeSimple(departure_time=3, arrival_time_target=4))
        self.assertEquals(4, node_profile.evaluate_earliest_arrival_time_at_target(2, 0))

    def test_pareto_optimality(self):
        node_profile = NodeProfileSimple()

        pair1 = LabelTimeSimple(departure_time=1, arrival_time_target=2)
        self.assertTrue(node_profile.update_pareto_optimal_tuples(pair1))

        pair2 = LabelTimeSimple(departure_time=2, arrival_time_target=3)
        self.assertTrue(node_profile.update_pareto_optimal_tuples(pair2))

        self.assertEquals(2, len(node_profile._labels))

        pair3 = LabelTimeSimple(departure_time=1, arrival_time_target=1)
        self.assertTrue(node_profile.update_pareto_optimal_tuples(pair3))
        self.assertEquals(2, len(node_profile._labels), msg=str(node_profile.get_final_optimal_labels()))

        pair4 = LabelTimeSimple(departure_time=1, arrival_time_target=2)
        self.assertFalse(node_profile.update_pareto_optimal_tuples(pair4))

    def test_pareto_optimality2(self):
        node_profile = NodeProfileSimple()
        pt2 = LabelTimeSimple(departure_time=10, arrival_time_target=35)
        self.assertTrue(node_profile.update_pareto_optimal_tuples(pt2))
        pt1 = LabelTimeSimple(departure_time=5, arrival_time_target=35)
        self.assertFalse(node_profile.update_pareto_optimal_tuples(pt1))
        self.assertEquals(len(node_profile.get_final_optimal_labels()), 1)

    def test_identity_profile(self):
        identity_profile = NodeProfileSimple(0)
        self.assertFalse(identity_profile.update_pareto_optimal_tuples(LabelTimeSimple(10, 10)))
        self.assertEqual(10, identity_profile.evaluate_earliest_arrival_time_at_target(10, 0))

    def test_walk_duration(self):
        node_profile = NodeProfileSimple(walk_to_target_duration=27)
        self.assertEqual(27, node_profile.get_walk_to_target_duration())
        pt1 = LabelTimeSimple(departure_time=5, arrival_time_target=35)
        self.assertFalse(node_profile.update_pareto_optimal_tuples(pt1))
        pt2 = LabelTimeSimple(departure_time=10, arrival_time_target=35)
        self.assertTrue(node_profile.update_pareto_optimal_tuples(pt2))

    def test_pareto_optimality_with_transfers(self):
        node_profile = NodeProfileSimple(label_class=LabelTimeWithBoardingsCount)
        pt3 = LabelTimeWithBoardingsCount(departure_time=5, arrival_time_target=35, n_boardings=0, first_leg_is_walk=True)
        pt2 = LabelTimeWithBoardingsCount(departure_time=5, arrival_time_target=35, n_boardings=1, first_leg_is_walk=True)
        pt1 = LabelTimeWithBoardingsCount(departure_time=5, arrival_time_target=35, n_boardings=2, first_leg_is_walk=True)
        self.assertTrue(node_profile.update_pareto_optimal_tuples(pt1))
        self.assertTrue(node_profile.update_pareto_optimal_tuples(pt2))
        self.assertTrue(node_profile.update_pareto_optimal_tuples(pt3))
        self.assertEqual(1, len(node_profile.get_final_optimal_labels()))




from gtfspy.routing.connection import Connection


def compute_pseudo_connections(transit_connections, start_time_dep,
                               end_time_dep, transfer_margin,
                               walk_network, walk_speed):
    """
    Given a set of transit events and the static walk network,
    "transform" the static walking network into a set of "pseudo-connections".

    As a first approximation, we add pseudo-connections to depart after each arrival of a transit connection
    to it's arrival stop.

    Parameters
    ----------
    transit_connections: list[Connection]
    start_time_dep : int
        start time in unixtime seconds
    end_time_dep: int
        end time in unixtime seconds (no new connections will be scanned after this time)
    transfer_margin: int
        required extra margin required for transfers in seconds
    walk_speed: float
        walking speed between stops in meters / second
    walk_network: networkx.Graph
        each edge should have the walking distance as a data attribute ("d_walk") expressed in meters

    Returns
    -------
    pseudo_connections: set[Connection]
    """
    # A pseudo-connection should be created after (each) arrival to a transit_connection's arrival stop.
    pseudo_connection_set = set()  # use a set to ignore possible duplicates
    for c in transit_connections:
        if start_time_dep <= c.departure_time <= end_time_dep:
            walk_arr_stop = c.departure_stop
            walk_arr_time = c.departure_time - transfer_margin
            for _, walk_dep_stop, data in walk_network.edges(nbunch=[walk_arr_stop], data=True):
                walk_dep_time = walk_arr_time - data['d_walk'] / float(walk_speed)
                if walk_dep_time > end_time_dep or walk_dep_time < start_time_dep:
                    continue
                pseudo_connection = Connection(walk_dep_stop,
                                               walk_arr_stop,
                                               walk_dep_time,
                                               walk_arr_time,
                                               Connection.WALK_TRIP_ID,
                                               Connection.WALK_SEQ,
                                               is_walk=True)
                pseudo_connection_set.add(pseudo_connection)
    return pseudo_connection_set






from __future__ import absolute_import, print_function

import numpy
import pandas as pd

from gtfspy.gtfs import GTFS
from .event import Event
from .heap import EventHeap
from .spreading_stop import SpreadingStop


class Spreader(object):
    """
    Starting from a specific point and time, get complete single source
    shortest path spreading dynamics as trips, or "events".
    """

    def __init__(self, gtfs, start_time_ut, lat, lon, max_duration_ut, min_transfer_time=30,
                 shapes=True, walk_speed=0.5):
        """
        Parameters
        ----------
        gtfs: GTFS
            the underlying GTFS (database) connection for getting data
        start_time_ut: number
            Start time of the spreading.
        lat: float
            latitude of the spreading seed location
        lon: float
            longitude of the spreading seed location
        max_duration_ut: int
            maximum duration of the spreading process (in seconds)
        min_transfer_time : int
            minimum transfer time in seconds
        shapes : bool
            whether to include shapes
        """
        self.gtfs = gtfs
        self.start_time_ut = start_time_ut
        self.lat = lat
        self.lon = lon
        self.max_duration_ut = max_duration_ut
        self.min_transfer_time = min_transfer_time
        self.shapes = shapes
        self.event_heap = None
        self.walk_speed = walk_speed
        self._uninfected_stops = None
        self._stop_I_to_spreading_stop = None
        self._initialized = False
        self._has_run = False

    def spread(self):
        self._initialize()
        self._run()
        return self._get_shortest_path_trips()

    def _initialize(self):
        if self._initialized:
            raise RuntimeError("This spreader instance has already been initialized: "
                               "create a new Spreader object for a new run.")
        # events are sorted by arrival time, so in order to use the
        # heapq, we need to have events coded as
        # (arrival_time, (from_stop, to_stop))
        start_stop_I = self.gtfs.get_closest_stop(self.lat, self.lon)
        end_time_ut = self.start_time_ut + self.max_duration_ut

        print("Computing/fetching events")
        events_df = self.gtfs.get_transit_events(self.start_time_ut, end_time_ut)
        all_stops = set(self.gtfs.stops()['stop_I'])

        self._uninfected_stops = all_stops.copy()
        self._uninfected_stops.remove(start_stop_I)

        # match stop_I to a more advanced stop object
        seed_stop = SpreadingStop(start_stop_I, self.min_transfer_time)

        self._stop_I_to_spreading_stop = {
            start_stop_I: seed_stop
        }
        for stop in self._uninfected_stops:
            self._stop_I_to_spreading_stop[stop] = SpreadingStop(stop, self.min_transfer_time)

        # get for each stop their
        print("intializing heap")
        self.event_heap = EventHeap(events_df)

        start_event = Event(self.start_time_ut - 1,
                            self.start_time_ut - 1,
                            start_stop_I,
                            start_stop_I,
                            -1)

        seed_stop.visit(start_event)
        assert len(seed_stop.visit_events) > 0
        self.event_heap.add_event(start_event)
        transfer_distances = self.gtfs.get_straight_line_transfer_distances(start_event.to_stop_I)
        self.event_heap.add_walk_events_to_heap(
            transfer_distances,
            start_event,
            self.start_time_ut,
            self.walk_speed,
            self._uninfected_stops,
            self.max_duration_ut
        )
        self._initialized = True

    def _run(self):
        """
        Run the actual simulation.
        """
        if self._has_run:
            raise RuntimeError("This spreader instance has already been run: "
                               "create a new Spreader object for a new run.")
        i = 1
        while self.event_heap.size() > 0 and len(self._uninfected_stops) > 0:
            event = self.event_heap.pop_next_event()
            this_stop = self._stop_I_to_spreading_stop[event.from_stop_I]

            if event.arr_time_ut > self.start_time_ut + self.max_duration_ut:
                break

            if this_stop.can_infect(event):

                target_stop = self._stop_I_to_spreading_stop[event.to_stop_I]
                already_visited = target_stop.has_been_visited()
                target_stop.visit(event)

                if not already_visited:
                    self._uninfected_stops.remove(event.to_stop_I)
                    print(i, self.event_heap.size())
                    transfer_distances = self.gtfs.get_straight_line_transfer_distances(event.to_stop_I)
                    self.event_heap.add_walk_events_to_heap(transfer_distances, event, self.start_time_ut,
                                                            self.walk_speed, self._uninfected_stops,
                                                            self.max_duration_ut)
                    i += 1
        self._has_run = True

    def _get_shortest_path_trips(self):
        """
        Returns
        -------
        trips: list
            trips['trips'] is a list whose each element (e.g. el = trips['trips'][0])
            is a dict with the following properties:
                el['lats'] : list of latitudes
                el['lons'] : list of longitudes
                el['times'] : list of passage_times
                el['route_type'] : type of vehicle as specified by GTFS (and -1 for walking)
                el['name'] : name of the route
        """
        if not self._has_run:
            raise RuntimeError("This spreader object has not run yet. Can not return any trips.")
        # create new transfer events and add them to the heap (=queue)
        inf_times = [[stop_I, el.get_min_visit_time() - self.start_time_ut]
                     for stop_I, el in self._stop_I_to_spreading_stop.items()]
        inf_times = numpy.array(inf_times)
        inf_time_data = pd.DataFrame(inf_times, columns=["stop_I", "inf_time_ut"])
        stop_data = self.gtfs.stops()

        combined = inf_time_data.merge(stop_data, how='inner', on='stop_I', suffixes=('_infs', '_stops'), copy=True)

        trips = []
        for stop_I, dest_stop_obj in self._stop_I_to_spreading_stop.items():
            inf_event = dest_stop_obj.get_min_event()
            if inf_event is None:
                continue
            dep_stop_I = inf_event.from_stop_I
            dep_lat = float(combined[combined['stop_I'] == dep_stop_I]['lat'].values)
            dep_lon = float(combined[combined['stop_I'] == dep_stop_I]['lon'].values)

            dest_lat = float(combined[combined['stop_I'] == stop_I]['lat'].values)
            dest_lon = float(combined[combined['stop_I'] == stop_I]['lon'].values)

            if inf_event.trip_I == -1:
                name = "walk"
                rtype = -1
            else:
                name, rtype = self.gtfs.get_route_name_and_type_of_tripI(inf_event.trip_I)

            trip = {
                "lats"      : [dep_lat, dest_lat],
                "lons"      : [dep_lon, dest_lon],
                "times"     : [inf_event.dep_time_ut, inf_event.arr_time_ut],
                "name"      : name,
                "route_type": rtype
            }
            trips.append(trip)
        return {"trips": trips}



from collections import Counter, defaultdict
import sys


class WarningsContainer(object):

    def __init__(self):
        self._warnings_counter = Counter()
        # key: "warning type" string, value: "number of errors" int
        self._warnings_records = defaultdict(list)
        # key: "warning_type" string, value: "list of dataframes etc." providing the details

    def add_warning(self, warning, reason, count=None):
        if count == 0:
            return
        if count is not None:
            self._warnings_counter[warning] += count
        else:
            self._warnings_counter[warning] += 1
        self._warnings_records[warning].append(reason)

    def write_summary(self, output_stream=None):
        if output_stream is None:
            output_stream = sys.stdout
        output_stream.write('The feed produced the following warnings:\n')
        for warning, count in self._warnings_counter.most_common():
            output_stream.write(warning + ": " + str(count) + "\n")

    def write_details(self, output_stream=None):
        if output_stream is None:
            output_stream = sys.stdout
        output_stream.write('The feed produced the following warnings (with details):\n')
        for warning, count in self._warnings_counter.most_common():
            output_stream.write(warning + ": " + str(count) + "\n")
            for reason in self._warnings_records[warning]:
                output_stream.write(str(reason) + "\n")

    def get_warning_counter(self):
        """
        Returns
        -------
        counter: collections.Counter
        """
        return self._warnings_counter

    def get_warnings_by_query_rows(self):
        """
        Returns
        -------
        warnings_record: defaultdict(list)
            maps each row to a list of warnings
        """
        return self._warnings_records

    def clear(self):
        self._warnings_counter.clear()
        self._warnings_records.clear()


from gtfspy.import_loaders.table_loader import TableLoader, decode_six


class ShapeLoader(TableLoader):
    fname = 'shapes.txt'
    table = 'shapes'
    tabledef = '(shape_id TEXT, lat REAL, lon REAL, seq INT, d INT)'

    # shape_id,shape_pt_lat,shape_pt_lon,shape_pt_sequence
    # 1001_20140811_1,60.167430,24.951684,1
    def gen_rows(self, readers, prefixes):
        for reader, prefix in zip(readers, prefixes):
            for row in reader:
                #print row
                yield dict(
                    shape_id      = prefix + decode_six(row['shape_id']),
                    lat           = float(row['shape_pt_lat']),
                    lon           = float(row['shape_pt_lon']),
                    seq           = int(row['shape_pt_sequence'])
                )

    @classmethod
    def index(cls, cur):
        # cur.execute('CREATE INDEX IF NOT EXISTS idx_shapes_shid ON shapes (shape_id)')
        # cur.execute('CREATE INDEX IF NOT EXISTS idx_shapes_id_seq ON shapes (shape_I, seq)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_shapes_id_seq ON shapes (shape_id, seq)')

    @classmethod
    def post_import(cls, cur):
        from gtfspy import shapes
        cur.execute('SELECT DISTINCT shape_id FROM shapes')
        shape_ids = tuple(x[0] for x in cur)

        # print "Renumbering sequences to start from 0 and Calculating shape cumulative distances"
        for shape_id in shape_ids:
            rows = cur.execute("SELECT shape_id, seq "
                               "FROM shapes "
                               "WHERE shape_id=? "
                               "ORDER BY seq", (shape_id,)
            ).fetchall()
            cur.executemany("UPDATE shapes SET seq=? "
                            "WHERE shape_id=? AND seq=?",
                            ( (i, shape_id, seq)
                             for i, (shape_id, seq) in enumerate(rows))
                            )

        for shape_id in shape_ids:
            shape_points = shapes.get_shape_points(cur, shape_id)
            shapes.gen_cumulative_distances(shape_points)

            cur.executemany('UPDATE shapes SET d=? '
                            'WHERE shape_id=? AND seq=? ',
                            ((pt['d'], shape_id, pt['seq'])
                             for pt in shape_points))


from gtfspy.import_loaders.agency_loader import AgencyLoader
from gtfspy.import_loaders.calendar_dates_loader import CalendarDatesLoader
from gtfspy.import_loaders.calendar_loader import CalendarLoader
from gtfspy.import_loaders.day_loader import DayLoader
from gtfspy.import_loaders.day_trips_materializer import DayTripsMaterializer
from gtfspy.import_loaders.feed_info_loader import FeedInfoLoader
from gtfspy.import_loaders.frequencies_loader import FrequenciesLoader
from gtfspy.import_loaders.trip_loader import TripLoader
from gtfspy.import_loaders.metadata_loader import MetadataLoader
from gtfspy.import_loaders.route_loader import RouteLoader
from gtfspy.import_loaders.shape_loader import ShapeLoader
from gtfspy.import_loaders.stop_distances_loader import StopDistancesLoader
from gtfspy.import_loaders.stop_loader import StopLoader
from gtfspy.import_loaders.stop_times_loader import StopTimesLoader
from gtfspy.import_loaders.transfer_loader import TransfersLoader